#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the SRT reader"""

# pylint: disable=R0201,C0115,C0116,W0212
import unittest
import io

from ttconv.srt.config import SRTReaderConfiguration
from ttconv.srt.reader import to_model
import ttconv.style_properties as styles
import ttconv.model as model



class SrtReaderTest(unittest.TestCase):

  def test_sample(self):
    # from https://en.wikipedia.org/wiki/SubRip
    SAMPLE = """1
00:02:16,612 --> 00:02:19,376
Senator, we're making
our final approach into Coruscant.

2
00:02:19,482 --> 00:02:21,609
Very good, Lieutenant.

3
00:03:13,336 --> 00:03:15,167
We made it.

4
00:03:18,608 --> 00:03:20,371
I guess I was wrong.

5
00:03:20,476 --> 00:03:22,671
There was no danger at all."""

    f = io.StringIO(SAMPLE)

    self.assertIsNotNone(to_model(f))

  def test_bold(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello <b>my</b> name is Bob
""")
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        break
    else:
      self.fail()

  def test_blank_lines(self):
    # from https://en.wikipedia.org/wiki/SubRip
    SAMPLE = """

1
00:02:16,612 --> 00:02:19,376
Senator, we're making
our final approach into Coruscant.


2
00:02:19,482 --> 00:02:21,609
Very good, Lieutenant.

5
00:03:20,476 --> 00:03:22,671
There was no danger at all.



"""

    f = io.StringIO(SAMPLE)

    self.assertIsNotNone(to_model(f))

  def test_bold_alt(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello {bold}my{/bold} name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        self.fail()
  
  def test_bold_alt2(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello <bold>my</bold> name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        self.fail()

  def test_bold_alt3(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello {b}my{/b} name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        self.fail()

  def test_italic(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello <i>my</i> name is Bob
""")
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontStyle) == styles.FontStyleType.italic:
        break
    else:
      self.fail()

  def test_italic_alt(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello {italic}my{/italic} name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontStyle) == styles.FontStyleType.italic:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontStyle) == styles.FontStyleType.italic:
        self.fail()

  def test_italic_alt1(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello {i}my{/i} name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontStyle) == styles.FontStyleType.italic:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontStyle) == styles.FontStyleType.italic:
        self.fail()

  def test_italic_alt2(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello <italic>my</italic> name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontStyle) == styles.FontStyleType.italic:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontStyle) == styles.FontStyleType.italic:
        self.fail()

  def test_underline(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello <u>my</u> name is Bob
""")
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      text_decoration = e.get_style(styles.StyleProperties.TextDecoration)
      if text_decoration is not None and text_decoration.underline:
        break
    else:
      self.fail()

  def test_underline_alt(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello {underline}my{/underline} name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      text_decoration = e.get_style(styles.StyleProperties.TextDecoration)
      if text_decoration is not None and text_decoration.underline:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      text_decoration = e.get_style(styles.StyleProperties.TextDecoration)
      if text_decoration is not None and text_decoration.underline:
        self.fail()

  def test_underline_alt1(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello {u}my{/u} name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      text_decoration = e.get_style(styles.StyleProperties.TextDecoration)
      if text_decoration is not None and text_decoration.underline:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      text_decoration = e.get_style(styles.StyleProperties.TextDecoration)
      if text_decoration is not None and text_decoration.underline:
        self.fail()

  def test_underline_alt2(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello <underline>my</underline> name is Bob
""")
    doc = to_model(f, SRTReaderConfiguration(extended_tags=True))
    for e in doc.get_body().dfs_iterator():
      text_decoration = e.get_style(styles.StyleProperties.TextDecoration)
      if text_decoration is not None and text_decoration.underline:
        break
    else:
      self.fail()
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      text_decoration = e.get_style(styles.StyleProperties.TextDecoration)
      if text_decoration is not None and text_decoration.underline:
        self.fail()

  def test_blue(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello <font color='blue'>my</font> name is Bob
""")

    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.Color) == styles.NamedColors.blue.value:
        break
    else:
      self.fail()

  def test_multiline_tags(self):
    f = io.StringIO(r"""1
00:02:16,612 --> 00:02:19,376
Hello <b>my
</b> name is Bob
""")
    doc = to_model(f)
    for e in doc.get_body().dfs_iterator():
      if e.get_style(styles.StyleProperties.FontWeight) == styles.FontWeightType.bold:
        break
    else:
      self.fail()

  def test_long_hours(self):
    f = io.StringIO(r"""1
101:00:00,000 --> 101:00:01,000
Hello <bold>my</bold> name is Bob
""")
    doc = to_model(f)

    self.assertEqual(
      363600,
      doc.get_body().first_child().first_child().get_begin()
    )

    self.assertEqual(
      363601,
      doc.get_body().first_child().first_child().get_end()
    )

  def test_single_line_text(self):
    f = io.StringIO(r"""1
101:00:00,000 --> 101:00:01,000
Hello
""")
    doc = to_model(f)

    p_children = list(doc.get_body().first_child().first_child())
    self.assertEqual(len(p_children), 1)
    self.assertIsInstance(p_children[0], model.Span)
    self.assertEqual(p_children[0].first_child().get_text(), "Hello")

  def test_multiline_text(self):
    f = io.StringIO(r"""1
101:00:00,000 --> 101:00:01,000
Hello
World
""")
    doc = to_model(f)

    p_children = list(doc.get_body().first_child().first_child())
    self.assertEqual(len(p_children), 3)
    self.assertIsInstance(p_children[0], model.Span)
    self.assertEqual(p_children[0].first_child().get_text(), "Hello")
    self.assertIsInstance(p_children[1], model.Br)
    self.assertIsInstance(p_children[2], model.Span)
    self.assertEqual(p_children[2].first_child().get_text(), "World")

  def test_alignment_tags_disabled_by_default(self):
    """Alignment tags should NOT be parsed when alignment_tags=False (default)"""
    f = io.StringIO(r"""1
00:00:00,000 --> 00:00:01,000
{\an1} Bottom Left
""")
    doc = to_model(f)
    # Should only have the default region (r_an2)
    regions = list(doc.iter_regions())
    self.assertEqual(len(regions), 1)
    self.assertEqual(regions[0].get_id(), "r_an2")
    # The alignment tag should remain in the text
    p = doc.get_body().first_child().first_child()
    text_content = ""
    for e in p.dfs_iterator():
      if isinstance(e, model.Text):
        text_content += e.get_text()
    self.assertIn("{\\an1}", text_content)

  def test_alignment_tags_all_positions(self):
    """Test all 9 alignment positions with alignment_tags=True"""
    f = io.StringIO(r"""1
00:00:00,000 --> 00:00:01,000
{\an1} V: Bottom - H: Left

2
00:00:02,000 --> 00:00:03,000
{\an2} V: Bottom - H: Center

3
00:00:04,000 --> 00:00:05,000
{\an3} V: Bottom - H: Right

4
00:00:06,000 --> 00:00:07,000
{\an4} V: Center - H: Left

5
00:00:08,000 --> 00:00:09,000
{\an5} V: Center - H: Center

6
00:00:10,000 --> 00:00:11,000
{\an6} V: Center - H: Right

7
00:00:12,000 --> 00:00:13,000
{\an7} V: Top - H: Left

8
00:00:14,000 --> 00:00:15,000
{\an8} V: Top - H: Center

9
00:00:16,000 --> 00:00:17,000
{\an9} V: Top - H: Right
""")
    doc = to_model(f, SRTReaderConfiguration(alignment_tags=True))
    
    # Should have 9 alignment regions (r_an2 is shared as default)
    regions = list(doc.iter_regions())
    self.assertEqual(len(regions), 9)  # r_an1 through r_an9
    
    # Check alignment regions exist with correct properties
    expected_alignments = {
      1: (styles.DisplayAlignType.after, styles.TextAlignType.start),
      2: (styles.DisplayAlignType.after, styles.TextAlignType.center),
      3: (styles.DisplayAlignType.after, styles.TextAlignType.end),
      4: (styles.DisplayAlignType.center, styles.TextAlignType.start),
      5: (styles.DisplayAlignType.center, styles.TextAlignType.center),
      6: (styles.DisplayAlignType.center, styles.TextAlignType.end),
      7: (styles.DisplayAlignType.before, styles.TextAlignType.start),
      8: (styles.DisplayAlignType.before, styles.TextAlignType.center),
      9: (styles.DisplayAlignType.before, styles.TextAlignType.end),
    }
    
    for code, (display_align, text_align) in expected_alignments.items():
      region = doc.get_region(f"r_an{code}")
      self.assertIsNotNone(region, f"Region r_an{code} should exist")
      self.assertEqual(
        region.get_style(styles.StyleProperties.DisplayAlign),
        display_align,
        f"r_an{code} DisplayAlign"
      )
      self.assertEqual(
        region.get_style(styles.StyleProperties.TextAlign),
        text_align,
        f"r_an{code} TextAlign"
      )

  def test_alignment_tag_stripped_from_text(self):
    """Alignment tag should be removed from displayed text"""
    f = io.StringIO(r"""1
00:00:00,000 --> 00:00:01,000
{\an7} Top Left Text
""")
    doc = to_model(f, SRTReaderConfiguration(alignment_tags=True))
    p = doc.get_body().first_child().first_child()
    text_content = ""
    for e in p.dfs_iterator():
      if isinstance(e, model.Text):
        text_content += e.get_text()
    self.assertNotIn("{\\an", text_content)
    self.assertIn("Top Left Text", text_content)

  def test_alignment_region_safe_area(self):
    """Alignment regions should use fixed 10% safe area margin"""
    f = io.StringIO(r"""1
00:00:00,000 --> 00:00:01,000
{\an1} Test
""")
    doc = to_model(f, SRTReaderConfiguration(alignment_tags=True))
    region = doc.get_region("r_an1")
    origin = region.get_style(styles.StyleProperties.Origin)
    extent = region.get_style(styles.StyleProperties.Extent)
    
    # Fixed 10% safe area: origin at (10%, 10%), extent is (80%, 80%)
    self.assertEqual(origin.x.value, 10)
    self.assertEqual(origin.y.value, 10)
    self.assertEqual(extent.width.value, 80)
    self.assertEqual(extent.height.value, 80)

  def test_alignment_mixed_paragraphs(self):
    """Mix of paragraphs with and without alignment tags"""
    f = io.StringIO(r"""1
00:00:00,000 --> 00:00:01,000
No alignment tag here

2
00:00:02,000 --> 00:00:03,000
{\an7} Top left aligned

3
00:00:04,000 --> 00:00:05,000
Also no alignment tag
""")
    doc = to_model(f, SRTReaderConfiguration(alignment_tags=True))
    
    paragraphs = list(doc.get_body().first_child())
    self.assertEqual(len(paragraphs), 3)
    
    # First and third paragraphs should not have a region set (use default)
    self.assertIsNone(paragraphs[0].get_region())
    self.assertIsNone(paragraphs[2].get_region())
    
    # Second paragraph should have r_an7 region
    self.assertEqual(paragraphs[1].get_region().get_id(), "r_an7")

  def test_alignment_region_reuse(self):
    """Multiple paragraphs with same alignment should share region"""
    f = io.StringIO(r"""1
00:00:00,000 --> 00:00:01,000
{\an1} First bottom left

2
00:00:02,000 --> 00:00:03,000
{\an1} Second bottom left

3
00:00:04,000 --> 00:00:05,000
{\an9} Top right
""")
    doc = to_model(f, SRTReaderConfiguration(alignment_tags=True))
    
    paragraphs = list(doc.get_body().first_child())
    
    # First two paragraphs should share the same region
    self.assertIs(paragraphs[0].get_region(), paragraphs[1].get_region())
    self.assertEqual(paragraphs[0].get_region().get_id(), "r_an1")
    
    # Third paragraph should have different region
    self.assertEqual(paragraphs[2].get_region().get_id(), "r_an9")
    
    # Total alignment regions should be 2 (r_an1 and r_an9) + default r_an2
    regions = list(doc.iter_regions())
    self.assertEqual(len(regions), 3)

  def test_alignment_multiple_tags_uses_first(self):
    """Multiple alignment tags in same caption: uses first, removes all"""
    f = io.StringIO(r"""1
00:00:00,000 --> 00:00:01,000
{\an1}{\an9} Multiple tags here
""")
    doc = to_model(f, SRTReaderConfiguration(alignment_tags=True))
    p = doc.get_body().first_child().first_child()
    
    # Should use first alignment (an1 = bottom-left)
    self.assertEqual(p.get_region().get_id(), "r_an1")
    
    # Both tags should be removed from text
    text_content = ""
    for e in p.dfs_iterator():
      if isinstance(e, model.Text):
        text_content += e.get_text()
    self.assertNotIn("{\\an", text_content)
    self.assertIn("Multiple tags here", text_content)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC Background and Foreground Attribute Codes"""

from __future__ import annotations

from typing import Optional

from ttconv.scc.codes import SccCode
from ttconv.style_properties import ColorType, NamedColors, TextDecorationType


class SccAttributeCode(SccCode):
  """SCC Foreground and Background Attribute Codes definition"""
  BWO = (0x1020, 0x1820, ColorType((0xFF, 0xFF, 0xFF, 0xFF)))  # Background White, Opaque
  BWS = (0x1021, 0x1821, ColorType((0xFF, 0xFF, 0xFF, 0x88)))  # Background White, Semi-transparent
  BGO = (0x1022, 0x1822, ColorType((0x00, 0xFF, 0x00, 0xFF)))  # Background Green, Opaque
  BGS = (0x1023, 0x1823, ColorType((0x00, 0xFF, 0x00, 0x88)))  # Background Green, Semi-transparent
  BBO = (0x1024, 0x1824, ColorType((0x00, 0x00, 0xFF, 0xFF)))  # Background Blue, Opaque
  BBS = (0x1025, 0x1825, ColorType((0x00, 0x00, 0xFF, 0x88)))  # Background Blue, Semi-transparent
  BCO = (0x1026, 0x1826, ColorType((0x00, 0xFF, 0xFF, 0xFF)))  # Background Cyan, Opaque
  BCS = (0x1027, 0x1827, ColorType((0x00, 0xFF, 0xFF, 0x88)))  # Background Cyan, Semi-transparent
  BRO = (0x1028, 0x1828, ColorType((0xFF, 0x00, 0x00, 0xFF)))  # Background Red, Opaque
  BRS = (0x1029, 0x1829, ColorType((0xFF, 0x00, 0x00, 0x88)))  # Background Red, Semi-transparent
  BYO = (0x102A, 0x182A, ColorType((0xFF, 0xFF, 0x00, 0xFF)))  # Background Yellow, Opaque
  BYS = (0x102B, 0x182B, ColorType((0xFF, 0xFF, 0x00, 0x88)))  # Background Yellow, Semi-transparent
  BMO = (0x102C, 0x182C, ColorType((0xFF, 0x00, 0xFF, 0xFF)))  # Background Magenta, Opaque
  BMS = (0x102D, 0x182D, ColorType((0xFF, 0x00, 0xFF, 0x88)))  # Background Magenta, Semi-transparent
  BAO = (0x102E, 0x182E, ColorType((0x00, 0x00, 0x00, 0xFF)))  # Background Black, Opaque
  BAS = (0x102F, 0x182F, ColorType((0x00, 0x00, 0x00, 0x88)))  # Background Black, Semi-transparent
  BT = (0x172D, 0x1F2D, ColorType((0x00, 0x00, 0x00, 0x00)))  # Background Transparent
  FA = (0x172E, 0x1F2E, NamedColors.black.value, False)  # Foreground Black
  FAU = (0x172F, 0x1F2F, NamedColors.black.value, False)  # Foreground Black Underline

  def __init__(self, channel_1: int, channel_2: int, color: ColorType, background=True):
    super().__init__(channel_1, channel_2)
    self._color = color
    self._background = background

  def get_color(self) -> ColorType:
    """Returns the corresponding color"""
    return self._color

  def get_name(self) -> str:
    """Returns the Attribute Code name"""
    return self.name

  def is_background(self) -> bool:
    """Returns whether the current code is a background or foreground code"""
    return self._background

  def get_text_decoration(self) -> Optional[TextDecorationType]:
    """Returns the corresponding text decoration"""
    if self is SccAttributeCode.FAU:
      return TextDecorationType(underline=True)
    return None

  @staticmethod
  def find(value: int) -> Optional[SccAttributeCode]:
    """Find the Attribute Code corresponding to the specified value"""
    for attribute_code in list(SccAttributeCode):
      if attribute_code.contains_value(value):
        return attribute_code
    return None

  def debug(self, value: int) -> str:
    """Debug representation of the code"""
    return "[" + str(self.get_channel(value)) + "|ATC|" + self.get_name() + "/" + hex(value) + "]"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2021, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""STL reader"""

from __future__ import annotations

import typing
import itertools
import logging

from ttconv.stl.config import STLReaderConfiguration
from ttconv.stl.datafile import DataFile

LOGGER = logging.getLogger(__name__)

#
# STL reader
#

def to_model(data_file: typing.IO, config: typing.Optional[STLReaderConfiguration] = None, progress_callback=lambda _: None):
  """Converts an STL document to the data model"""

  m = DataFile(
    data_file.read(1024),
    disable_fill_line_gap=False if config is None else config.disable_fill_line_gap,
    disable_line_padding=False if config is None else config.disable_line_padding,
    start_tc=None if config is None else config.program_start_tc,
    font_stack=None if config is None else config.font_stack,
    max_row_count=None if config is None else config.max_row_count
    )

  for i in itertools.count():

    buf = data_file.read(128)

    if not buf:
      break

    try:
      m.process_tti_block(buf)
    except:
      LOGGER.error("Bad TTI block")
      raise
    
    progress_callback(i/m.get_tti_count())

  return m.get_document()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2021, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""STL text field (TF) processing"""

import logging
import codecs

from ttconv import model
import ttconv.style_properties as styles
from ttconv.stl import iso6937

LOGGER = logging.getLogger(__name__)

_CHAR_DECODER_MAP = {
  b'00': iso6937.decode,
  b'01': codecs.getdecoder("iso8859_5"),
  b'02' : codecs.getdecoder("iso8859_6"),
  b'03': codecs.getdecoder("iso8859_7"),
  b'04':codecs.getdecoder("iso8859_8")
}

_UNUSED_SPACE_CODE = 0x8F

def _is_character_code(c) -> bool:
  return 0x20 <= c <= 0x7F or 0xA0 <= c <= 0xFF

def _is_printable_code(c) -> bool:
  return _is_character_code(c) and c != 0x20
 
def _is_control_code(c) -> bool:
  return 0x00 <= c <= 0x07 or 0x0A <= c <= 0x0D or 0x1C <= c <= 0x1D or 0x80 <= c <= 0x85

def _is_newline_code(c) -> bool:
  return c == 0x8A

def _is_unused_space_code(c) -> bool:
  return c == _UNUSED_SPACE_CODE

def _is_space_code(c) -> bool:
  return c == 0x20

def _is_lwsp_code(c) -> bool:
  return _is_control_code(c) or _is_newline_code(c) or _is_space_code(c)

def _note_decode_error(error):
  LOGGER.warning("Unknown character: %s", hex(error.object[error.start]))
  return ("�", error.end)

codecs.register_error("note", _note_decode_error)

def line_count(text_field: bytes, is_double_height: bool) -> int:
  """Returns the number of lines separated by one or more newline characters
  in the TF field `text_field`
  """
  count = 0
  was_eol = False

  for c in text_field:
    if _is_newline_code(c):
      if is_double_height:
        if was_eol:
          was_eol = False
        else:
          count += 1
          was_eol = True
      else:
        count += 1
        was_eol = True
    else:
      was_eol= False

  return count + 1

def has_double_height_char(text_field) -> bool:
  """Returns true if TF field `text_field` contains any double-height characters"""
  return any(map(lambda c: c == 0x0D, text_field))

class _TextFieldIterator:

  def __init__(self, text_field):
    self.pos = 0
    self.tf = text_field

  def __iter__(self):
    return self

  def peek_next(self) -> int:
    return _UNUSED_SPACE_CODE if self.pos + 1 >= len(self.tf) else self.tf[self.pos + 1]

  def peek_prev(self) -> int:
    return _UNUSED_SPACE_CODE if self.pos <= 0 else self.tf[self.pos - 1]

  def cur(self) -> int:
    return _UNUSED_SPACE_CODE if self.pos >= len(self.tf) else self.tf[self.pos]

  def __next__(self) -> int:
    c = self.peek_next()
    self.pos = min(self.pos + 1, len(self.tf))
    return c

class _Context:

  def __init__(self, parent_element: model.ContentElement, is_teletext: bool, decode_func):
    self.fg_color = None
    self.bg_color = None
    self.is_italic = None
    self.is_underline = None
    self.reset_styles(is_teletext)

    self.span = None
    self.parent = parent_element
    self.decode_func = decode_func
    self.text_buffer = bytearray()

  def reset_styles(self, is_teletext: bool):
    self.fg_color = styles.NamedColors.white.value
    self.bg_color = styles.NamedColors.black.value if is_teletext else styles.NamedColors.transparent.value
    self.is_italic = False
    self.is_underline = False

  def set_bg_color(self, color: styles.ColorType):
    self.bg_color = color 

  def get_bg_color(self) -> styles.ColorType:
    return self.bg_color

  def set_fg_color(self, color: styles.ColorType):
    self.fg_color = color 

  def get_fg_color(self) -> styles.ColorType:
    return self.fg_color

  def set_underline(self, is_underline: bool):
    self.is_underline = is_underline

  def get_underline(self) -> bool:
    return self.is_underline

  def set_italic(self, is_italic: bool):
    self.is_italic = is_italic

  def get_italic(self) -> bool:
    return self.is_italic

  def start_span(self):
    if self.span is None:
      self.span = model.Span(self.parent.get_doc())

      self.span.set_style(styles.StyleProperties.BackgroundColor, self.get_bg_color())

      self.span.set_style(styles.StyleProperties.Color, self.get_fg_color())

      if self.get_underline():
        self.span.set_style(
          styles.StyleProperties.TextDecoration,
          styles.TextDecorationType(underline=True)
        )

      if self.get_italic():
        self.span.set_style(
          styles.StyleProperties.FontStyle,
          styles.FontStyleType.italic
        )

  def end_span(self):
    if len(self.text_buffer) > 0 and self.span is not None:
      text_element = model.Text(self.parent.get_doc())
      text_element.set_text(self.decode_func(self.text_buffer, errors="note")[0])

      self.span.push_child(text_element)
      self.parent.push_child(self.span)
      self.span = None
      self.text_buffer.clear()

  def append_character(self, c):
    self.start_span()
    self.text_buffer.append(c)


def to_model(element: model.ContentElement, is_teletext: bool, tti_cct: bytes, tti_tf: bytes):
  """Converts the EBU STL text field `tti_tf` to a sequence of span elements and 
  appends them to the content element `element`. `is_teletext` indicates whether the text
  field is interpreted as teletext or open/undefined. `tti_cct` specifies the text field code page as
  signaled in the TTI CCT field.
  """

  tf_iter = _TextFieldIterator(tti_tf)

  decode_func = _CHAR_DECODER_MAP.get(tti_cct)

  if decode_func is None:
    decode_func = iso6937.decode
    LOGGER.error("Unknown Text Field character set: %s", str(tti_cct))

  context = _Context(element, is_teletext, decode_func)

  while True:

    c = tf_iter.cur()

    if _is_unused_space_code(c):
      break

    if _is_character_code(c):
      if _is_printable_code(c) or (_is_printable_code(tf_iter.peek_next()) and _is_printable_code(tf_iter.peek_prev())):
        context.append_character(c)

    elif _is_newline_code(c):
      if not _is_newline_code(tf_iter.peek_next()) and not _is_unused_space_code(tf_iter.peek_next()):
        context.end_span()
        element.push_child(model.Br(element.get_doc()))
        if is_teletext:
          context.reset_styles(is_teletext)

    elif _is_control_code(c):
      context.end_span()

      if c == 0x1C:
        context.set_bg_color(styles.NamedColors.black.value)
      elif c == 0x85:
        context.set_bg_color(styles.NamedColors.transparent.value)
      elif c == 0x1D:
        context.set_bg_color(context.get_fg_color())
      elif c == 0x00:
        context.set_fg_color(styles.NamedColors.black.value)
      elif c == 0x01:
        context.set_fg_color(styles.NamedColors.red.value)
      elif c == 0x02:
        context.set_fg_color(styles.NamedColors.lime.value)
      elif c == 0x03:
        context.set_fg_color(styles.NamedColors.yellow.value)
      elif c == 0x04:
        context.set_fg_color(styles.NamedColors.blue.value)
      elif c == 0x05:
        context.set_fg_color(styles.NamedColors.magenta.value)
      elif c == 0x06:
        context.set_fg_color(styles.NamedColors.cyan.value)
      elif c == 0x07:
        context.set_fg_color(styles.NamedColors.white.value)
      elif c == 0x80:
        context.set_italic(True)
      elif c == 0x81:
        context.set_italic(False)
      elif c == 0x82:
        context.set_underline(True)
      elif c == 0x83:
        context.set_underline(False)

      if (_is_printable_code(tf_iter.peek_next()) and _is_printable_code(tf_iter.peek_prev())):
        context.append_character(0X20)

    next(tf_iter)

  context.end_span()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the IMSC writer'''

# pylint: disable=R0201,C0115,C0116

import os
import re
import unittest
import logging
import json
from fractions import Fraction
import xml.etree.ElementTree as et
import xml.dom.minidom as minidom
import ttconv.imsc.reader as imsc_reader
import ttconv.imsc.writer as imsc_writer
import ttconv.model as model
import ttconv.style_properties as styles
import ttconv.imsc.namespaces as xml_ns
import ttconv.imsc.style_properties as imsc_styles
import ttconv.imsc.attributes as attributes
from ttconv.scc.utils import get_extent_from_dimensions
import ttconv.imsc.config as imsc_config

def _get_set_style(imsc_style_prop, model_value):
  e = et.Element("p")
  assert imsc_style_prop.model_prop.validate(model_value)
  imsc_style_prop.from_model(e, model_value)
  return e.attrib.get(f"{{{imsc_style_prop.ns}}}{imsc_style_prop.local_name}")

def _get_time(imsc_style_prop, model_value):
  e = et.Element("p")
  assert imsc_style_prop.model_prop.validate(model_value)
  imsc_style_prop.from_model(e, model_value)
  return e.attrib.get(f"{{{imsc_style_prop.ns}}}{imsc_style_prop.local_name}")

class ReaderWriterTest(unittest.TestCase):

  def setUp(self):
    if not os.path.exists('build'):
      os.makedirs('build')

  def write_pretty_xml(self, tree: et.ElementTree, file_path):
    rough_string = et.tostring(tree.getroot(), 'utf-8')
    reparsed = minidom.parseString(rough_string)
    with open(file_path, "wb") as f:
      f.write(reparsed.toprettyxml(indent="  ", encoding="utf-8"))

  def test_default_ns(self):
    """Confirm that the tt ns is the default namespace"""
    file_to_parse = "src/test/resources/ttml/imsc-tests/imsc1/ttml/animation/Animation001.ttml"
    tree = et.parse(file_to_parse)
    test_model = imsc_reader.to_model(tree)
    tree_from_model = imsc_writer.from_model(test_model)
    rough_string = et.tostring(tree_from_model.getroot(), 'unicode')

    self.assertNotRegex(rough_string, "ttml:tt")
    self.assertRegex(rough_string, "<tt")


  def test_animation_001(self):
    file_to_parse = "src/test/resources/ttml/imsc-tests/imsc1/ttml/animation/Animation001.ttml"
    
    tree = et.parse(file_to_parse)

    # create the model
    test_model = imsc_reader.to_model(tree)

    # convert from a model to a ttml document
    tree_from_model = imsc_writer.from_model(test_model)

    # write the document out to a file
    self.write_pretty_xml(tree_from_model, 'build/out.ttml')

  def test_imsc_1_test_suite(self):
    manifest = []
    base_path = "build/imsc1"

    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name), self.assertLogs() as logs:
            logging.getLogger().info("*****dummy*****") # dummy log
            tree = et.parse(os.path.join(root, filename))
            test_model = imsc_reader.to_model(tree)
            tree_from_model = imsc_writer.from_model(test_model)

            test_dir_relative_path = os.path.basename(root)

            test_relative_path = os.path.join(test_dir_relative_path, filename)

            os.makedirs(os.path.join(base_path, "ttml", test_dir_relative_path), exist_ok=True)

            with open(os.path.join(base_path, "ttml", test_relative_path), "wb") as f:
              f.write(et.tostring(tree_from_model.getroot(), 'utf-8'))

            manifest.append({"path" : str(test_relative_path).replace('\\', '/')})

            if len(logs.output) > 1:
              self.fail(logs.output)

    with open(os.path.join(base_path, "tests.json"), "w", encoding="utf8") as fp:
      json.dump(manifest, fp)

  def test_imsc_1_1_test_suite(self):
    manifest = []
    base_path = "build/imsc1_1"

    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1_1/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name), self.assertLogs() as logs:
            logging.getLogger().info("*****dummy*****") # dummy log
            tree = et.parse(os.path.join(root, filename))
            test_model = imsc_reader.to_model(tree)
            tree_from_model = imsc_writer.from_model(test_model)

            test_dir_relative_path = os.path.basename(root)

            test_relative_path = os.path.join(test_dir_relative_path, filename)

            os.makedirs(os.path.join(base_path, "ttml", test_dir_relative_path), exist_ok=True)

            with open(os.path.join(base_path, "ttml", test_relative_path), "wb") as f:
              f.write(et.tostring(tree_from_model.getroot(), 'utf-8'))

            manifest.append({"path" : str(test_relative_path).replace('\\', '/')})

            if len(logs.output) > 1:
              self.fail(logs.output)

    with open(os.path.join(base_path, "tests.json"), "w", encoding="utf8") as fp:
      json.dump(manifest, fp)

  def test_imsc_1_3_test_suite(self):
    manifest = []
    base_path = "build/imsc1_3"

    for root, _subdirs, files in os.walk("src/test/resources/ttml/imsc-tests/imsc1_3/ttml"):
      for filename in files:
        (name, ext) = os.path.splitext(filename)
        if ext == ".ttml":
          with self.subTest(name), self.assertLogs() as logs:
            logging.getLogger().info("*****dummy*****") # dummy log
            tree = et.parse(os.path.join(root, filename))
            test_model = imsc_reader.to_model(tree)
            tree_from_model = imsc_writer.from_model(test_model)

            test_dir_relative_path = os.path.basename(root)

            test_relative_path = os.path.join(test_dir_relative_path, filename)

            os.makedirs(os.path.join(base_path, "ttml", test_dir_relative_path), exist_ok=True)

            with open(os.path.join(base_path, "ttml", test_relative_path), "wb") as f:
              f.write(et.tostring(tree_from_model.getroot(), 'utf-8'))

            manifest.append({"path" : str(test_relative_path).replace('\\', '/')})

            if len(logs.output) > 1:
              self.fail(logs.output)

    with open(os.path.join(base_path, "tests.json"), "w", encoding="utf8") as fp:
      json.dump(manifest, fp)


class FromModelBodyWriterTest(unittest.TestCase):

  def setUp(self):
    if not os.path.exists('build'):
      os.makedirs('build')
    
    et.register_namespace("ttml", xml_ns.TTML)
    et.register_namespace("ttp", xml_ns.TTP)
    et.register_namespace("tts", xml_ns.TTS)
    et.register_namespace("ittp", xml_ns.ITTP)
    et.register_namespace("itts", xml_ns.ITTS)

  def test_body_only(self):

    doc = model.ContentDocument()
    body = model.Body(doc)
    div = model.Div(doc)
    p = model.P(doc)
    span = model.Span(doc)
    text = model.Text(doc)
    text.set_text("asdf")
    
    span.push_child(text)
    p.push_child(span)
    div.push_child(p)
    body.push_child(div)
    doc.set_body(body)

    # write the document out to a file
    imsc_writer.from_model(doc).write('build/BodyElement.out.ttml', encoding='utf-8', xml_declaration=True)

class StylePropertyWriterTest(unittest.TestCase):

  def test_tts_luminance_gain(self):
    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.LuminanceGain, 1.2),
      "1.2"
    )

  def test_tts_opacity(self):
    self.assertEqual(_get_set_style(imsc_styles.StyleProperties.Opacity, 0.532), "0.532")

  def test_tts_overflow(self):
    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.Overflow, styles.OverflowType.visible),
      "visible"
    )

  def test_tts_padding(self):
    padding = styles.PaddingType(
      before=styles.LengthType(10.1),
      end=styles.LengthType(20.2),
      after=styles.LengthType(30.3),
      start=styles.LengthType(40.4)
    )

    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.Padding, padding),
      r"10.1% 20.2% 30.3% 40.4%"
    )

  def test_tts_ruby_align(self):
    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.RubyAlign, styles.RubyAlignType.spaceAround),
      "spaceAround"
    )

  def test_tts_ruby_position(self):
    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.RubyPosition, styles.AnnotationPositionType.outside),
      "outside"
    )

  def test_tts_ruby_reserve(self):
    rr1 = styles.SpecialValues.none
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.RubyReserve,
        rr1
        ),
      "none"
    )

    rr2 = styles.RubyReserveType(
      position=styles.RubyReserveType.Position.both,
      length=styles.LengthType(value=1.2, units=styles.LengthType.Units.em)
    )
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.RubyReserve,
        rr2
        ),
      "both 1.2em"
    )

    rr3 = styles.RubyReserveType(
      position=styles.RubyReserveType.Position.outside,
    )
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.RubyReserve,
        rr3
        ),
      "outside"
    )

  def test_tts_shear(self):
    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.Shear, 12.3),
      "12.3%"
    )

  def test_tts_text_align(self):
    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.TextAlign, styles.TextAlignType.start),
      "start"
    )

  def test_tts_text_combine(self):
    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.TextCombine, styles.TextCombineType.all),
      "all"
    )

  def test_tts_text_decoration(self):
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.TextDecoration,
        styles.TextDecorationType(
          underline=False,
          line_through=True
        )
      ),
      "noUnderline lineThrough"
    )

  def test_tts_text_emphasis(self):
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.TextEmphasis,
        styles.TextEmphasisType(
          style=styles.TextEmphasisType.Style.filled_circle,
          color=styles.NamedColors.red.value,
          position=styles.TextEmphasisType.Position.after
        )
      ),
      "filled circle #ff0000 after"
    )

  def test_tts_text_outline(self):
    to1 = styles.SpecialValues.none
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.TextOutline,
        to1
        ),
      "none"
    )

    to2 = styles.TextOutlineType(
      color=styles.NamedColors.red.value,
      thickness=styles.LengthType(value=5)
    )
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.TextOutline,
        to2
        ),
      "#ff0000 5%"
    )

  def test_tts_text_shadow(self):
    ts1 = styles.SpecialValues.none
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.TextShadow,
        ts1
        ),
      "none"
    )

    ts2 = styles.TextShadowType(
      (
        styles.TextShadowType.Shadow(
          x_offset=styles.LengthType(value=1, units=styles.LengthType.Units.em),
          y_offset=styles.LengthType(value=1.2, units=styles.LengthType.Units.em)
        ),
        styles.TextShadowType.Shadow(
          x_offset=styles.LengthType(value=0.5, units=styles.LengthType.Units.em),
          y_offset=styles.LengthType(value=0.7, units=styles.LengthType.Units.em),
          blur_radius=styles.LengthType(value=1, units=styles.LengthType.Units.em),
          color=styles.NamedColors.red.value
        )
      )
    )

    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.TextShadow,
        ts2
        ),
      "1em 1.2em, 0.5em 0.7em 1em #ff0000"
    )

  def test_tts_unicode_bidi(self):
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.UnicodeBidi,
        styles.UnicodeBidiType.bidiOverride
      ),
      "bidiOverride"
    )

  def test_tts_visibility(self):
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.Visibility,
        styles.VisibilityType.hidden
      ),
      "hidden"
    )

  def test_tts_wrap_option(self):
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.WrapOption,
        styles.WrapOptionType.noWrap
      ),
      "noWrap"
    )

  def test_tts_writing_mode(self):
    self.assertEqual(
      _get_set_style(
        imsc_styles.StyleProperties.WritingMode,
        styles.WritingModeType.tbrl
      ),
      "tbrl"
    )

  def test_tts_writing_extent(self):
    self.assertEqual(
      _get_set_style(imsc_styles.StyleProperties.Extent,
      get_extent_from_dimensions(123, 456, styles.LengthType.Units.px)),
      '123px 456px'
    )

  def test_tts_writing_no_extent_when_no_body(self):

    d = model.ContentDocument()

    tree_from_model = imsc_writer.from_model(d)

    extent = tree_from_model.getroot().attrib.get(
      f"{{{imsc_styles.StyleProperties.Extent.ns}}}{imsc_styles.StyleProperties.Extent.local_name}")
    
    self.assertEqual(extent, None)
  
  def test_tts_writing_no_extent_when_body_has_no_extents(self):

    doc = model.ContentDocument()
    body = model.Body(doc)
    div = model.Div(doc)
    p = model.P(doc)
    span = model.Span(doc)
    text = model.Text(doc)
    text.set_text("asdf")
    
    span.push_child(text)
    p.push_child(span)
    div.push_child(p)
    body.push_child(div)
    doc.set_body(body)

    tree_from_model = imsc_writer.from_model(doc)

    extent = tree_from_model.getroot().attrib.get(
      f"{{{imsc_styles.StyleProperties.Extent.ns}}}{imsc_styles.StyleProperties.Extent.local_name}")
    
    self.assertEqual(extent, None)

  def test_tts_writing_extent_when_body_has_extents(self):

    doc = model.ContentDocument()
    body = model.Body(doc)
    div = model.Div(doc)

    div.set_style(styles.StyleProperties.Extent,
      get_extent_from_dimensions(123, 456, styles.LengthType.Units.px))

    p = model.P(doc)
    span = model.Span(doc)
    text = model.Text(doc)
    text.set_text("asdf")
    
    span.push_child(text)
    p.push_child(span)
    div.push_child(p)
    body.push_child(div)
    doc.set_body(body)

    r = model.Region("hello", doc)
    r.set_style(styles.StyleProperties.Extent,
      get_extent_from_dimensions(123, 456, styles.LengthType.Units.px))

    doc.put_region(r)

    tree_from_model = imsc_writer.from_model(doc)

    extent = tree_from_model.getroot().attrib.get(
      f"{{{imsc_styles.StyleProperties.Extent.ns}}}{imsc_styles.StyleProperties.Extent.local_name}")
    
    self.assertEqual(extent, '1920px 1080px')

  def test_style_property_base_has_px(self):
    prop = imsc_styles.StyleProperty
    self.assertEqual(imsc_styles.StyleProperty.has_px(prop),
      False
    )

  def test_style_property_disparity_has_px(self):
    prop = styles.LengthType(1, styles.LengthType.units.em)
    self.assertEqual(imsc_styles.StyleProperties.Disparity.has_px(prop), False)

    prop = styles.LengthType(1, styles.LengthType.units.px)
    self.assertEqual(imsc_styles.StyleProperties.Disparity.has_px(prop), True)

  def test_style_property_extent_has_px(self):
    prop = styles.ExtentType(styles.LengthType(1, styles.LengthType.units.px), 
      styles.LengthType(1, styles.LengthType.units.em))
    self.assertEqual(imsc_styles.StyleProperties.Extent.has_px(prop), True)

    prop = styles.ExtentType(styles.LengthType(1, styles.LengthType.units.px), 
      styles.LengthType(1, styles.LengthType.units.px))
    self.assertEqual(imsc_styles.StyleProperties.Extent.has_px(prop), True)

    prop = styles.ExtentType(styles.LengthType(1, styles.LengthType.units.em), 
      styles.LengthType(1, styles.LengthType.units.px))
    self.assertEqual(imsc_styles.StyleProperties.Extent.has_px(prop), True)

    prop = styles.ExtentType(styles.LengthType(1, styles.LengthType.units.em), 
      styles.LengthType(1, styles.LengthType.units.em))
    self.assertEqual(imsc_styles.StyleProperties.Extent.has_px(prop), False)

  def test_style_property_font_size_has_px(self):
    prop = styles.LengthType(1, styles.LengthType.units.em)
    self.assertEqual(imsc_styles.StyleProperties.FontSize.has_px(prop), False)

    prop = styles.LengthType(1, styles.LengthType.units.px)
    self.assertEqual(imsc_styles.StyleProperties.FontSize.has_px(prop), True)

  def test_style_property_line_height_has_px(self):
    prop = styles.SpecialValues.normal
    self.assertEqual(imsc_styles.StyleProperties.LineHeight.has_px(prop), False)

    prop = styles.LengthType(1, styles.LengthType.units.em)
    self.assertEqual(imsc_styles.StyleProperties.LineHeight.has_px(prop), False)

    prop = styles.LengthType(1, styles.LengthType.units.px)
    self.assertEqual(imsc_styles.StyleProperties.LineHeight.has_px(prop), True)

  def test_style_property_origin_has_px(self):
    prop = styles.CoordinateType(styles.LengthType(1, styles.LengthType.units.px), 
      styles.LengthType(1, styles.LengthType.units.em))
    self.assertEqual(imsc_styles.StyleProperties.Origin.has_px(prop), True)

    prop = styles.CoordinateType(styles.LengthType(1, styles.LengthType.units.px), 
      styles.LengthType(1, styles.LengthType.units.px))
    self.assertEqual(imsc_styles.StyleProperties.Origin.has_px(prop), True)

    prop = styles.CoordinateType(styles.LengthType(1, styles.LengthType.units.em), 
      styles.LengthType(1, styles.LengthType.units.px))
    self.assertEqual(imsc_styles.StyleProperties.Origin.has_px(prop), True)

    prop = styles.CoordinateType(styles.LengthType(1, styles.LengthType.units.em), 
      styles.LengthType(1, styles.LengthType.units.em))
    self.assertEqual(imsc_styles.StyleProperties.Origin.has_px(prop), False)

  def test_style_property_padding_has_px(self):
    prop = styles.PaddingType(
      before = styles.LengthType(10.1, styles.LengthType.units.px),
      end = styles.LengthType(20.2, styles.LengthType.units.em),
      after = styles.LengthType(30.3, styles.LengthType.units.em),
      start = styles.LengthType(40.4, styles.LengthType.units.em)
    )
    self.assertEqual(imsc_styles.StyleProperties.Padding.has_px(prop), True)

    prop = styles.PaddingType(
      before = styles.LengthType(10.1, styles.LengthType.units.em),
      end = styles.LengthType(20.2, styles.LengthType.units.px),
      after = styles.LengthType(30.3, styles.LengthType.units.em),
      start = styles.LengthType(40.4, styles.LengthType.units.em)
    )
    self.assertEqual(imsc_styles.StyleProperties.Padding.has_px(prop), True)

    prop = styles.PaddingType(
      before = styles.LengthType(10.1, styles.LengthType.units.em),
      end = styles.LengthType(20.2, styles.LengthType.units.em),
      after = styles.LengthType(30.3, styles.LengthType.units.px),
      start = styles.LengthType(40.4, styles.LengthType.units.em)
    )
    self.assertEqual(imsc_styles.StyleProperties.Padding.has_px(prop), True)

    prop = styles.PaddingType(
      before = styles.LengthType(10.1, styles.LengthType.units.em),
      end = styles.LengthType(20.2, styles.LengthType.units.em),
      after = styles.LengthType(30.3, styles.LengthType.units.em),
      start = styles.LengthType(40.4, styles.LengthType.units.px)
    )
    self.assertEqual(imsc_styles.StyleProperties.Padding.has_px(prop), True)

    prop = styles.PaddingType(
      before = styles.LengthType(10.1, styles.LengthType.units.em),
      end = styles.LengthType(20.2, styles.LengthType.units.em),
      after = styles.LengthType(30.3, styles.LengthType.units.em),
      start = styles.LengthType(40.4, styles.LengthType.units.em)
    )
    self.assertEqual(imsc_styles.StyleProperties.Padding.has_px(prop), False)

  def test_style_property_ruby_reserve_has_px(self):
    prop = styles.RubyReserveType(
      position=styles.RubyReserveType.Position.both,
      length=styles.LengthType(value=1.2, units=styles.LengthType.Units.px)
    )
    self.assertEqual(imsc_styles.StyleProperties.RubyReserve.has_px(prop), True)

    prop = styles.RubyReserveType(
      position=styles.RubyReserveType.Position.both,
      length=styles.LengthType(value=1.2, units=styles.LengthType.Units.em)
    )
    self.assertEqual(imsc_styles.StyleProperties.RubyReserve.has_px(prop), False)


  def test_style_property_text_outline_has_px(self):
    prop = styles.SpecialValues.none
    self.assertEqual(imsc_styles.StyleProperties.TextOutline.has_px(prop), False)

    prop = styles.TextOutlineType(
      color=styles.NamedColors.red.value,
      thickness=styles.LengthType(value=5, units=styles.LengthType.Units.em)
    )
    self.assertEqual(imsc_styles.StyleProperties.TextOutline.has_px(prop), False)

    prop = styles.TextOutlineType(
      color=styles.NamedColors.red.value,
      thickness=styles.LengthType(value=5, units=styles.LengthType.Units.px)
    )
    self.assertEqual(imsc_styles.StyleProperties.TextOutline.has_px(prop), True)
    
  def test_style_property_text_shadow_has_px(self):
    prop = styles.TextShadowType(
      (
        styles.TextShadowType.Shadow(
          x_offset=styles.LengthType(value=1, units=styles.LengthType.Units.em),
          y_offset=styles.LengthType(value=1, units=styles.LengthType.Units.em)
        ),
        styles.TextShadowType.Shadow(
          x_offset=styles.LengthType(value=0.5, units=styles.LengthType.Units.em),
          y_offset=styles.LengthType(value=0.7, units=styles.LengthType.Units.em),
          blur_radius=styles.LengthType(value=1, units=styles.LengthType.Units.em),
          color=styles.NamedColors.red.value
        )
      )
    )
    self.assertEqual(imsc_styles.StyleProperties.TextShadow.has_px(prop), False)

    prop = styles.TextShadowType(
      (
        styles.TextShadowType.Shadow(
          x_offset=styles.LengthType(value=1, units=styles.LengthType.Units.px),
          y_offset=styles.LengthType(value=1.2, units=styles.LengthType.Units.em)
        ),
        styles.TextShadowType.Shadow(
          x_offset=styles.LengthType(value=0.5, units=styles.LengthType.Units.em),
          y_offset=styles.LengthType(value=0.7, units=styles.LengthType.Units.em),
          blur_radius=styles.LengthType(value=1, units=styles.LengthType.Units.em),
          color=styles.NamedColors.red.value
        )
      )
    )
    self.assertEqual(imsc_styles.StyleProperties.TextShadow.has_px(prop), True)

    prop = styles.TextShadowType(
      (
        styles.TextShadowType.Shadow(
          x_offset=styles.LengthType(value=1, units=styles.LengthType.Units.em),
          y_offset=styles.LengthType(value=1.2, units=styles.LengthType.Units.em)
        ),
        styles.TextShadowType.Shadow(
          x_offset=styles.LengthType(value=0.5, units=styles.LengthType.Units.px),
          y_offset=styles.LengthType(value=0.7, units=styles.LengthType.Units.em),
          blur_radius=styles.LengthType(value=1, units=styles.LengthType.Units.em),
          color=styles.NamedColors.red.value
        )
      )
    )
    self.assertEqual(imsc_styles.StyleProperties.TextShadow.has_px(prop), True)

class WriterWithTimeFormattingTest(unittest.TestCase):

  def test_write_with_frames(self):

    context = attributes.TemporalAttributeWritingContext(None)
    time = Fraction(1, 4)
    val = attributes.to_time_format(context, time)
    self.assertEqual(val, '00:00:00.250')

    context = attributes.TemporalAttributeWritingContext(None)
    time = Fraction(1, 3)
    val = attributes.to_time_format(context, time)
    self.assertEqual(val, '00:00:00.333')

    context = attributes.TemporalAttributeWritingContext(Fraction(30000, 1000))
    time = Fraction(1, 3)
    val = attributes.to_time_format(context, time)
    self.assertEqual(val, '00:00:00.333')

    context = attributes.TemporalAttributeWritingContext(Fraction(30000, 1000), attributes.TimeExpressionSyntaxEnum.frames)
    time = Fraction(2, 1)
    val = attributes.to_time_format(context, time)
    self.assertEqual(val, '60f')

    context = attributes.TemporalAttributeWritingContext(Fraction(30000, 1001), attributes.TimeExpressionSyntaxEnum.frames)
    time = Fraction(3, 1)
    val = attributes.to_time_format(context, time)
    self.assertEqual(val, '90f')

    context = attributes.TemporalAttributeWritingContext(Fraction(25000, 1000), attributes.TimeExpressionSyntaxEnum.frames)
    time = Fraction(3600, 1)
    val = attributes.to_time_format(context, time)
    self.assertEqual(val, '90000f')

    context = attributes.TemporalAttributeWritingContext(Fraction(30000, 1001), attributes.TimeExpressionSyntaxEnum.frames)
    time = Fraction(0.7674333333333333)
    val = attributes.to_time_format(context, time)
    self.assertEqual(val, '23f')

    context = attributes.TemporalAttributeWritingContext(Fraction(30000, 1001), attributes.TimeExpressionSyntaxEnum.frames)
    time = 2301 / Fraction(30000, 1001)
    val = attributes.to_time_format(context, time)
    self.assertEqual(val, '2301f')
    

class ConfigTest(unittest.TestCase):

  def test_profile_signaling(self):

    source_cps = {"http://www.w3.org/ns/ttml/profile/imsc1.1/text", "http://www.w3.org/ns/ttml/profile/imsc1/text"}

    ttml_doc = model.ContentDocument()
    ttml_doc.set_content_profiles(source_cps)

    config = imsc_config.IMSCWriterConfiguration.parse({"profile_signaling" : "content_profiles"})
    xml_from_model = imsc_writer.from_model(ttml_doc, config)
    extracted_cps = {e for e in re.findall(r'\S+', xml_from_model.getroot().get(f"{{{xml_ns.TTP}}}contentProfiles"))}
    self.assertSetEqual(extracted_cps, source_cps)

    xml_from_model = imsc_writer.from_model(ttml_doc, None)
    self.assertIsNone(xml_from_model.getroot().get(f"{{{xml_ns.TTP}}}contentProfiles"))

    config = imsc_config.IMSCWriterConfiguration.parse({"profile_signaling" : "none"})
    xml_from_model = imsc_writer.from_model(ttml_doc, config)
    self.assertIsNone(xml_from_model.getroot().get(f"{{{xml_ns.TTP}}}contentProfiles"))

  def test_clock_time(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
    <tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml" >
    <body begin="2.3s"/>
    </tt>
    """

    ttml_doc = et.ElementTree(et.fromstring(ttml_doc_str))
    
    config = imsc_config.IMSCWriterConfiguration.parse({"time_format" : "clock_time"})
    xml_from_model = imsc_writer.from_model(imsc_reader.to_model(ttml_doc), config)
    body_element = xml_from_model.find("tt:body", {"tt": xml_ns.TTML})
    self.assertEqual(body_element.get("begin"), "00:00:02.300")

    xml_from_model = imsc_writer.from_model(imsc_reader.to_model(ttml_doc))
    body_element = xml_from_model.find("tt:body", {"tt": xml_ns.TTML})
    self.assertEqual(body_element.get("begin"), "00:00:02.300")


  def test_frames(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
    <tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml" >
    <body begin="5s"/>
    </tt>
    """

    ttml_doc = et.ElementTree(et.fromstring(ttml_doc_str))
    
    config = imsc_config.IMSCWriterConfiguration.parse({"time_format": "frames", "fps": "30/1"})
    xml_from_model = imsc_writer.from_model(imsc_reader.to_model(ttml_doc), config)
    body_element = xml_from_model.find("tt:body", {"tt": xml_ns.TTML})
    self.assertEqual(body_element.get("begin"), "150f")

    config = imsc_config.IMSCWriterConfiguration.parse({"fps": "30/1"})
    xml_from_model = imsc_writer.from_model(imsc_reader.to_model(ttml_doc), config)
    body_element = xml_from_model.find("tt:body", {"tt": xml_ns.TTML})
    self.assertEqual(body_element.get("begin"), "150f")

    config = imsc_config.IMSCWriterConfiguration.parse({"time_format": "frames"})
    with self.assertRaises(ValueError):
      imsc_writer.from_model(imsc_reader.to_model(ttml_doc), config)
    
  def test_clock_time_with_frames(self):
    ttml_doc_str = """<?xml version="1.0" encoding="UTF-8"?>
    <tt xml:lang="en" xmlns="http://www.w3.org/ns/ttml" >
    <body begin="2.1s"/>
    </tt>
    """

    ttml_doc = et.ElementTree(et.fromstring(ttml_doc_str))
    
    config = imsc_config.IMSCWriterConfiguration.parse({"time_format": "clock_time_with_frames", "fps": "30/1"})
    xml_from_model = imsc_writer.from_model(imsc_reader.to_model(ttml_doc), config)
    body_element = xml_from_model.find("tt:body", {"tt": xml_ns.TTML})
    self.assertEqual(body_element.get("begin"), "00:00:02:03")

    config = imsc_config.IMSCWriterConfiguration.parse({"time_format": "clock_time_with_frames"})
    with self.assertRaises(ValueError):
      imsc_writer.from_model(imsc_reader.to_model(ttml_doc), config)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2022, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""VTT reader"""

from __future__ import annotations

from dataclasses import dataclass
import typing
import re
import logging
from enum import Enum

from ttconv import model
from ttconv import style_properties as styles
from ttconv.vtt.tokenizer import EndTagToken, StartTagToken, StringToken, CueTextTokenizer, TimestampTagToken, Token

LOGGER = logging.getLogger(__name__)

def _none_terminated(iterator):
  for item in iterator:
    yield item
  yield None

def _parse_cue_text(cue_text: str, paragraph: model.P, line_number: int):
  parser = _TextCueParser(paragraph, line_number)

  for token in CueTextTokenizer(cue_text):
    parser.handle_token(token)

class _TextCueParser:

  def __init__(self, paragraph: model.P, line_number: int) -> None:
    self.line_num: int = line_number
    self.paragraph: model.P = paragraph
    self.parent: model.ContentElement = paragraph

    # handle the special case of ruby elements where children cannot be added one by one
    self.ruby_rbc: typing.Optional[model.Rbc] = None
    self.ruby_rtc: typing.Optional[model.Rtc] = None

  def handle_token(self, token: Token) -> None:
    if isinstance(token, StartTagToken):
      self._handle_starttag(token)
    elif isinstance(token, EndTagToken):
      self._handle_endtag(token)
    elif isinstance(token, StringToken):
      self._handle_string(token)
    elif isinstance(token, TimestampTagToken):
      self._handle_ts(token)
    else:
      raise ValueError("Unknown token type")

  def _handle_ts(self, token: TimestampTagToken):

    ts = vtt_timestamp_to_secs(token.timestamp)
    if ts is None:
      LOGGER.warning("Invalid timestamp tag %s at line %s", token.timestamp, self.line_num)
      return

    # we handle only top-level timestamp tags
    if self.parent.get_begin() is None:
      LOGGER.warning("Nested timestamp tag %s at line %s", token.timestamp, self.line_num)
      return

    p_begin = self.paragraph.get_begin()
    if p_begin is None or p_begin >= ts:
      LOGGER.warning("Invalid timestamp tag %s", token.timestamp)

    span = self._make_span(self.paragraph)
    span.set_begin(ts - p_begin)
    self.paragraph.push_child(span)
    self.parent = span

  def _handle_starttag(self, token: StartTagToken):

    tag = token.tag.lower()

    if tag.startswith("ruby"):
      if self.ruby_rbc is not None or self.ruby_rtc is not None:
        raise RuntimeError("Nested ruby tags are not allowed.")
      span = model.Ruby(self.parent.get_doc())

      # wrap <rb> and <rt> into <rbc> and <rtc>

      self.ruby_rbc = model.Rbc(self.parent.get_doc())
      self.ruby_rtc = model.Rtc(self.parent.get_doc())
      span.push_children([self.ruby_rbc, self.ruby_rtc])
      self.parent.push_child(span)
      self.parent = span
      return

    if tag.startswith("rt"):
      span = model.Rt(self.parent.get_doc())
      self.ruby_rtc.push_child(span)
      self.parent = span
      return

    # all other tags can be handled as a span

    span = self._make_span(self.parent)
    self.parent.push_child(span)
    self.parent = span

    if isinstance(span.parent(), model.P):
      span.set_style(styles.StyleProperties.BackgroundColor, _DEFAULT_BG_COLOR)

    if tag.startswith("b"):
      span.set_style(styles.StyleProperties.FontWeight, styles.FontWeightType.bold)

    elif tag.startswith("i"):
      span.set_style(styles.StyleProperties.FontStyle, styles.FontStyleType.italic)

    elif tag.startswith("u"):
      span.set_style(styles.StyleProperties.TextDecoration, styles.TextDecorationType(underline=True))

    elif tag.startswith("lang"):
      try:
        span.set_lang(token.annotation)
      except IndexError:
        LOGGER.warning("Lang tag without language present")

    elif tag.startswith("c"):
      if token.classes is not None:
        for c in token.classes:
          try:
            if c.startswith("bg_"):
              bg_color =  styles.NamedColors[c[3:]]
              span.set_style(styles.StyleProperties.BackgroundColor, bg_color.value)
            else:
              color =  styles.NamedColors[c]
              span.set_style(styles.StyleProperties.Color, color.value)
          except KeyError:
            LOGGER.warning("Ignoring class %s", c)

    elif tag == "v":
      pass

    else:
      LOGGER.warning("Unknown tag %s at line %s", tag, self.line_num)
      return

  def _handle_endtag(self, _token: EndTagToken):

    if isinstance(self.parent, model.Ruby):
      self.ruby_rbc = None
      self.ruby_rtc = None
    elif isinstance(self.parent, (model.Rt, model.Rb)):
      # this is needed since <rb> and <rt> are nested in <rbc> and <rtc>
      self.parent = self.parent.parent()

    self.parent = self.parent.parent()

  def _handle_string(self, token: StringToken):
    lines = token.value.split("\n")

    for i, line in enumerate(lines):
      if i > 0:
        self.parent.push_child(model.Br(self.parent.get_doc()))
      span = self._make_span(self.parent)
      span.push_child(model.Text(self.parent.get_doc(), line))
      if isinstance(self.parent, model.Ruby):
        rb = model.Rb(self.parent.get_doc())
        rb.push_child(span)
        self.ruby_rbc.push_child(rb)
      else:
        self.parent.push_child(span)

  def _make_span(self, parent: model.ContentElement) -> model.Span:
    span = model.Span(self.parent.get_doc())
    if isinstance(parent, model.P):
      span.set_style(styles.StyleProperties.BackgroundColor, _DEFAULT_BG_COLOR)
    return span


_EMPTY_RE = re.compile(r"[\n\r]*")
_DEFAULT_FONT_STACK = (styles.GenericFontFamilyType.sansSerif,)
_DEFAULT_FONT_SIZE_PCT = 5
_DEFAULT_FONT_SIZE = styles.LengthType(15 * _DEFAULT_FONT_SIZE_PCT, styles.LengthType.Units.pct) # 5vh for ttp:cellResolution="32 15"
_DEFAULT_TEXT_COLOR = styles.NamedColors.white.value
_DEFAULT_LINE_PADDING = styles.LengthType(0.5, styles.LengthType.Units.c)
_DEFAULT_LINE_HEIGHT = styles.LengthType(125, styles.LengthType.Units.pct)
_DEFAULT_LINE_HEIGHT_PCT = _DEFAULT_FONT_SIZE_PCT *_DEFAULT_LINE_HEIGHT.value/100
_DEFAULT_BG_COLOR = styles.ColorType((0, 0, 0, 204))
_DEFAULT_ROWS = 23
_DEFAULT_COLS = 40

_VTT_PCT_RE = re.compile(r"(\d+\.?\d*)%")

def parse_vtt_pct(value: str):
  """Parse a WebVTT precentage value"""
  m = _VTT_PCT_RE.fullmatch(value)
  if m:
    v = round(float(m.group(1)))
    return v if 0 <= v <= 100 else None
  return None

# integer has at most 20 digits
_VTT_INT_RE = re.compile(r"(-?\d{1,20})")

def parse_vtt_int(value: str):
  """Parse a WebVTT integer value"""
  m = _VTT_INT_RE.fullmatch(value)
  if m:
    return int(m.group(1))
  return None

@dataclass
class _RegionParams:
  extent : typing.Optional[styles.ExtentType] = None
  origin : typing.Optional[styles.CoordinateType] = None
  writing_mode : typing.Optional[styles.WritingModeType] = None
  text_align : typing.Optional[styles.TextAlignType] = None
  display_align : typing.Optional[styles.DisplayAlignType] = None
  
def make_region_params(cue_settings: dict[str, str]) -> _RegionParams:
  """Compute region style parameters from cue settings
  """
  r = _RegionParams()

  # writing_mode
  r.writing_mode = styles.WritingModeType.lrtb
  value = cue_settings.get("vertical")
  if value == "lr":
    r.writing_mode = styles.WritingModeType.tblr
  elif value == "rl":
    r.writing_mode = styles.WritingModeType.tbrl
  elif value is not None:
    LOGGER.warning("Bad vertical setting value: %s", value)

  # cue text align
  cue_text_align = "center"
  value = cue_settings.get("align")
  if value in ("left", "right", "start", "center", "end"):
    cue_text_align = value
  elif value is not None:
    LOGGER.warning("Bad align setting value: %s", cue_text_align)

  # cue size

  cue_size = 100
  value = cue_settings.get("size")
  if value is not None:
    value = parse_vtt_pct(value)
    if value is not None:
      cue_size = value
    else:
      LOGGER.warning("Bad size setting value: %s", value)

  # line setting

  cue_line_align = "start"
  cue_snap_to_lines = True
  cue_line_num = -1

  value = cue_settings.get("line")
  if value is not None:
    value = value.split(",")

    v = parse_vtt_pct(value[0])
    if v is not None:
      cue_line_num = v
      cue_snap_to_lines = False
    else:
      v = parse_vtt_int(value[0])
      if v is not None:
        cue_line_num = v
      else:
        LOGGER.warning("Bad line setting value: %s", value)

    if len(value) == 2:
      if value[1] in ("start", "center", "end"):
        cue_line_align = value[1]
      else:
        LOGGER.warning("Bad line alignment setting value: %s", value[1])

  # position

  cue_position = None
  cue_position_align = None

  value = cue_settings.get("position")
  if value is not None:
    value = value.split(",")

    # position value
    cue_position = parse_vtt_pct(value[0])
    if cue_position is None:
      LOGGER.warning("Bad position alignment setting value: %s", cue_position)
    else:
      # position alignment
      if len(value) > 1:
        if value[1] in ("center", "line-left", "line-right"):
          cue_position_align = value[1]
        else:
          LOGGER.warning("Bad position alignment setting value: %s", value[1])

  if cue_position_align is None:
    if cue_text_align == "start":
      cue_position_align = "line-right" if r.writing_mode == styles.WritingModeType.rltb else "line-left"
    elif cue_text_align == "end":
      cue_position_align = "line-left" if r.writing_mode == styles.WritingModeType.rltb else "line-right"
    elif cue_text_align == "left":
      cue_position_align = "line-left"
    elif cue_text_align == "right":
      cue_position_align = "line-right"
    elif cue_text_align == "center":
      cue_position_align = "center"
    else:
      raise RuntimeError("Invalid cue text alignment setting")  

  if cue_position is None or not (0 <= cue_position <= 100):
    if cue_text_align == "left":
      cue_position = 0
    elif cue_text_align == "right":
      cue_position = 100
    else:
      cue_position = 50

  # text_align

  r.text_align = styles.TextAlignType.center
  if cue_text_align == "left":
    r.text_align = styles.TextAlignType.end if r.writing_mode == styles.WritingModeType.rltb else styles.TextAlignType.start
  elif cue_text_align == "right":
    r.text_align = styles.TextAlignType.start if r.writing_mode == styles.WritingModeType.rltb else styles.TextAlignType.end
  elif cue_text_align == "start":
    r.text_align = styles.TextAlignType.start
  elif cue_text_align == "center":
    r.text_align = styles.TextAlignType.center
  elif cue_text_align == "end":
    r.text_align = styles.TextAlignType.end
  else:
    raise RuntimeError("Invalid cue text alignment setting")
  
  # display_align

  if cue_line_align == "center":
    r.display_align = styles.DisplayAlignType.center
  elif cue_line_align == "end":
    r.display_align = styles.DisplayAlignType.after
  elif cue_line_align == "start":
    r.display_align = styles.DisplayAlignType.before
  else:
    raise RuntimeError("Invalid line alignment setting")

  # compute region origin and extent

  if cue_position_align == "line-left":
    max_ipd_size = 100 - cue_position
  elif cue_position_align == "line-right":
    max_ipd_size = cue_position
  elif cue_position_align == "center":
    if cue_position <= 50:
      max_ipd_size = cue_position * 2
    else:
      max_ipd_size = (100 - cue_position) * 2
  else:
    raise RuntimeError("Invalid position alignment")

  ipd_size = max(_DEFAULT_FONT_SIZE_PCT, min(cue_size, max_ipd_size))

  if cue_position_align == "line-left":
    ipd_origin = min(cue_position, 100 - ipd_size)
  elif cue_position_align == "line-right":
    ipd_origin = max(0, cue_position - ipd_size)
  elif cue_position_align == "center":
    ipd_origin = max(min(cue_position - ipd_size / 2, 100 - ipd_size), 0)
  else:
    raise RuntimeError("Invalid position alignment")

  if cue_snap_to_lines:
    if cue_line_num >= 0:
      bpd_origin = _DEFAULT_LINE_HEIGHT_PCT * cue_line_num
    else:
      bpd_origin = 100 + cue_line_num * _DEFAULT_LINE_HEIGHT_PCT
  else:
    bpd_origin = cue_line_num

  if cue_line_align == "center":
    bpd_size = max(_DEFAULT_LINE_HEIGHT_PCT, min(bpd_origin * 2, (100 - bpd_origin) * 2))
  elif cue_line_align == "end":
    bpd_size = bpd_origin
  elif cue_line_align == "start":
    bpd_size = 100 - bpd_origin
  
  bpd_size = max(_DEFAULT_LINE_HEIGHT_PCT, bpd_size)

  if cue_line_align == "center":
    bpd_origin = max(min(bpd_origin - bpd_size / 2, 100 - bpd_size), 0)
  elif cue_line_align == "end":
    bpd_origin = max(0, bpd_origin - bpd_size)
  elif cue_line_align == "start":
    bpd_origin = min(bpd_origin, 100 - bpd_size)

  if r.writing_mode in (styles.WritingModeType.lrtb, styles.WritingModeType.rltb):
    r.extent = styles.ExtentType(
      height=styles.LengthType(bpd_size),
      width=styles.LengthType(ipd_size)
    )
    r.origin = styles.CoordinateType(
      x=styles.LengthType(ipd_origin),
      y=styles.LengthType(bpd_origin)
    )
  else:
    r.extent = styles.ExtentType(
      height=styles.LengthType(ipd_size),
      width=styles.LengthType(bpd_size)
    )
    r.origin = styles.CoordinateType(
      x=styles.LengthType(bpd_origin),
      y=styles.LengthType(ipd_origin)
    )
  
  return r

def _get_or_make_region(
  doc: model.ContentDocument,
  cue_settings_list: str
  ):
  """Returns a matching region from `doc` or creates one
  """

  cue_settings = dict(filter(lambda x: len(x) == 2, [x.split(":") for x in cue_settings_list]))

  if len(cue_settings) > 0:
    p = make_region_params(cue_settings)
  else:
    # maintain compatibility with earlier versions
    p = _RegionParams(
      extent = styles.ExtentType(
        height=styles.LengthType(100 - 200/_DEFAULT_ROWS),
        width=styles.LengthType(100 - 200/_DEFAULT_COLS)
      ),
      origin = styles.CoordinateType(
        x=styles.LengthType(100/_DEFAULT_COLS),
        y=styles.LengthType(100/_DEFAULT_ROWS)
      ),
      writing_mode=styles.WritingModeType.lrtb,
      text_align=styles.TextAlignType.center,
      display_align=styles.DisplayAlignType.after
    )

  # look for a matching region

  found_region = None
  regions = list(doc.iter_regions())
  for r in regions:

    if r.get_style(styles.StyleProperties.WritingMode) != p.writing_mode:
      continue

    if r.get_style(styles.StyleProperties.Extent) != p.extent:
      continue

    if r.get_style(styles.StyleProperties.Origin) != p.origin:
      continue

    if r.get_style(styles.StyleProperties.TextAlign) != p.text_align:
      continue

    if r.get_style(styles.StyleProperties.DisplayAlign) != p.display_align:
      continue

    found_region = r
    break

  if found_region is None:
    found_region = model.Region(f"r{len(regions)}", doc)
    found_region.set_style(styles.StyleProperties.Origin, p.origin)
    found_region.set_style(styles.StyleProperties.Extent, p.extent)
    found_region.set_style(styles.StyleProperties.DisplayAlign, p.display_align)
    found_region.set_style(styles.StyleProperties.TextAlign, p.text_align)
    found_region.set_style(styles.StyleProperties.WritingMode, p.writing_mode)
    found_region.set_style(styles.StyleProperties.LineHeight, _DEFAULT_LINE_HEIGHT)
    found_region.set_style(styles.StyleProperties.FontFamily, _DEFAULT_FONT_STACK)
    found_region.set_style(styles.StyleProperties.FontSize, _DEFAULT_FONT_SIZE)
    found_region.set_style(styles.StyleProperties.Color, _DEFAULT_TEXT_COLOR)
    found_region.set_style(styles.StyleProperties.LinePadding, _DEFAULT_LINE_PADDING)
    found_region.set_style(styles.StyleProperties.FillLineGap, True)
    doc.put_region(found_region)

  return found_region

_VTT_TS_RE = re.compile(r"(?:(?P<hh>[0-9]{2,}):)?(?P<mm>[0-9]{2}):(?P<ss>[0-9]{2})\.(?P<ms>[0-9]{3})")
_VTT_TS_TAG_RE = re.compile(r"<((?:[0-9]{2,3}:)?[0-9]{2}:[0-9]{2}\.[0-9]{3})>")

def vtt_timestamp_to_secs(vtt_ts: str):
  m = _VTT_TS_RE.fullmatch(vtt_ts)

  if m:
    return int(m.group('hh') if m.group('hh') is not None else 0) * 3600 + \
      int(m.group('mm')) * 60 + \
      int(m.group('ss')) + \
      int(m.group('ms')) / 1000

  return None

def to_model(data_file: typing.IO, _config = None, progress_callback=lambda _: None):
  """Converts a WebVTT document to the data model"""

  class _State(Enum):
    LOOKING = 1
    TC = 2
    TEXT = 3
    TEXT_MORE = 4
    START = 5
    NOTE = 6
    STYLE = 7


  doc = model.ContentDocument()

  body = model.Body(doc)
  doc.set_body(body)

  div = model.Div(doc)
  body.push_child(div)

  lines : str = data_file.readlines()

  state = _State.START
  current_p = None

  for line_index, line in enumerate(_none_terminated(lines)):

    if state is _State.START:
      if not line.startswith("WEBVTT"):
        LOGGER.warning("The first line of the file does not start with WEBVTT")
      state = _State.LOOKING
      continue

    if state in (_State.NOTE, _State.STYLE):
      # we skip over notes

      if line is None:
        break

      if _EMPTY_RE.fullmatch(line):
        state = _State.LOOKING
        continue

    if state is _State.LOOKING:
      if line is None:
        break

      if _EMPTY_RE.fullmatch(line):
        continue

      if line.startswith("NOTE "):
        state = _State.NOTE
        continue

      if line.startswith("STYLE"):
        state = _State.STYLE
        continue

      if "-->" not in line:
        # skip over cue id
        continue

      progress_callback(line_index/len(lines))

      cue_params = line.split()

      if len(cue_params) < 3:
        LOGGER.warning("Invalid line %s", line_index)
        continue

      start_time = vtt_timestamp_to_secs(cue_params[0])
      if start_time is None:
        LOGGER.warning("Invalid timestamp %s at %s", cue_params[0], line_index)
        continue

      end_time = vtt_timestamp_to_secs(cue_params[2])
      if end_time is None:
        LOGGER.warning("Invalid timestamp %s at %s", cue_params[2], line_index)
        continue

      current_p = model.P(doc)

      current_p.set_begin(start_time)

      current_p.set_end(end_time)

      # handle settings

      current_p.set_region(_get_or_make_region(doc, cue_params[3:]))

      state = _State.TEXT
      subtitle_text = None
      continue

    if state in (_State.TEXT, _State.TEXT_MORE):

      if line is None or _EMPTY_RE.fullmatch(line):
        if subtitle_text is not None:
          _parse_cue_text(
            subtitle_text.strip('\r\n').replace(r"\n\r", "\n"),
            current_p,
            line_index
          )
        else:
          LOGGER.warning("Ignoring cue due to a spurious blank line at line %s", line_index)

        state = _State.LOOKING
        continue

      if state is _State.TEXT:
        div.push_child(current_p)
        subtitle_text = ""

      subtitle_text += line

      state = _State.TEXT_MORE

      continue

  return doc



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2023, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Data model filter"""

from __future__ import annotations
from typing import Optional

from ttconv.model import ContentDocument
from ttconv.config import ModuleConfiguration

class DocumentFilter:
  """Abstract base class for content document filters"""

  _all_filters = dict()

  def __init__(self, config: ModuleConfiguration) -> None:
    self.config = config

  def process(self, doc: ContentDocument):
    """Processes the specified document in place."""
    raise NotImplementedError

  @classmethod
  def get_config_class(cls) -> ModuleConfiguration:
    """Returns the configuration class for the filter."""
    raise NotImplementedError

  @classmethod
  def get_filter_by_name(cls, name) -> Optional[DocumentFilter]:
    """Returns a list of all document filters"""
    return DocumentFilter._all_filters.get(name)

  def __init_subclass__(cls):
    DocumentFilter._all_filters[cls.get_config_class().name()] = cls

from ttconv.filters.doc import *



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2025, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC writer"""

from __future__ import annotations

import logging
from fractions import Fraction
import re
from typing import List, Optional, Sequence

import ttconv.model as model
from ttconv.isd import ISD
from ttconv.scc.codes.characters import unicode_to_scc
from ttconv.scc.codes.preambles_address_codes import SccPreambleAddressCode
from ttconv.scc.config import SccWriterConfiguration
from ttconv.style_properties import StyleProperties, NamedColors, TextAlignType
from ttconv.scc.codes.control_codes import SccControlCode
from ttconv.time_code import FPS_29_97, FPS_30, SmpteTimeCode

LOGGER = logging.getLogger(__name__)

class _Caption:
  """Collection of lines of text with a begin and end time and an alignment"""

  def __init__(self):
    self._begin: Fraction = 0
    self._end: Optional[Fraction] = None
    self._lines: List[str] = []
    self._alignment: TextAlignType = TextAlignType.start

  def set_begin(self, begin: Fraction):
    """Sets the begin time of the lines"""
    self._begin = begin

  def get_begin(self) -> Fraction:
    """Returns the begin time of the lines"""
    return self._begin

  def set_end(self, end: Fraction):
    """Sets the end time of the lines"""
    self._end = end

  def get_end(self) -> Optional[Fraction]:
    """Returns the end time of the lines"""
    return self._end

  def set_alignment(self, alignment: TextAlignType):
    """Sets the alignment of the lines"""
    self._alignment = alignment

  def get_alignment(self) -> TextAlignType:
    """Returns the alignment of the lines"""
    return self._alignment

  def set_lines(self, lines: List[str]):
    """Sets the lines of text"""
    self._lines = lines

  def __len__(self):
    return len(self._lines)

  def __getitem__(self, index: int) -> str:
    return self._lines[index]

  def __setitem__(self, index: int, value: str):
    self._lines[index] = value

  def __iter__(self):
    return iter(self._lines)

  def append(self, line: str):
    self._lines.append(line)

  @staticmethod
  def from_regions(regions: Sequence[model.Region]) -> _Caption:
    """Returns the list of lines that are flowed into the Region"""
    _lines: _Caption = _Caption()

    def _process_element(element: model.ContentElement):

      if isinstance(element, model.Div):
        for elem in element:
          _process_element(elem)
        return

      if isinstance(element, model.P):
        _lines.append("")
        _lines.set_alignment(element.get_style(StyleProperties.TextAlign))
        for elem in element:
          _process_element(elem)
        return

      if isinstance(element, model.Span):
        for elem in element:
          _process_element(elem)
        return

      if isinstance(element, model.Br):
        _lines.append("")
        return

      if isinstance(element, model.Text):
        _lines[-1] = _lines[-1] + element.get_text()
        return

    for region in regions:
      for body in region:
        for div in body:
          _process_element(div)

    return _lines

class _Chunk:

  ODD_PARITY_OCTETS = [
    128, 1, 2, 131, 4, 133, 134, 7, 8, 137, 138, 11, 140, 13, 14, 143,
    16, 145, 146, 19, 148, 21, 22, 151, 152, 25, 26, 155, 28, 157, 158, 31,
    32, 161, 162, 35, 164, 37, 38, 167, 168, 41, 42, 171, 44, 173, 174, 47,
    176, 49, 50, 179, 52, 181, 182, 55, 56, 185, 186, 59, 188, 61, 62, 191,
    64, 193, 194, 67, 196, 69, 70, 199, 200, 73, 74, 203, 76, 205, 206, 79,
    208, 81, 82, 211, 84, 213, 214, 87, 88, 217, 218, 91, 220, 93, 94, 223,
    224, 97, 98, 227, 100, 229, 230, 103, 104, 233, 234, 107, 236, 109, 110, 239,
    112, 241, 242, 115, 244, 117, 118, 247, 248, 121, 122, 251, 124, 253, 254, 127
  ]

  def __init__(self):
    self._octet_buffer : bytearray = bytearray()
    self._begin_frame : Optional[int] = None

  def set_begin(self, frame: Optional[int]):
    self._begin_frame = frame

  def get_begin(self) -> Optional[int]:
    return self._begin_frame

  def get_dur(self) -> int:
    return (len(self._octet_buffer) + 1) // 2

  def get_end(self) -> Optional[int]:
    if self._begin_frame is None:
      return None
    return self._begin_frame + self.get_dur()

  def push_control_code(self, cc: int):
    hi_octet = cc // 256
    lo_octet = cc % 256
    if (hi_octet > 127 or lo_octet > 127):
      raise RuntimeError("Line 21 octet is out of range")
    if len(self._octet_buffer) % 2 == 1:
      self._octet_buffer.append(0)
    self._octet_buffer.append(hi_octet)
    self._octet_buffer.append(lo_octet)

    # per spec, always double up control codes
    if 0x10 <= hi_octet <= 0x1F:
      self._octet_buffer.append(hi_octet)
      self._octet_buffer.append(lo_octet)

  def push_char(self, char: str):
    if len(char) != 1:
      raise ValueError("Length of string must be exactly 1")
    
    scc_bytes = unicode_to_scc(char)

    if len(scc_bytes) > 1 and len(self._octet_buffer) % 2 == 1:
      self._octet_buffer.append(0)
    
    for b in scc_bytes:
      self._octet_buffer.append(b)

  def overlap(self, other: _Chunk) -> bool:
    """Checks if the other chunk overlaps with the current chunk"""
    return other.get_begin() <= self.get_end() and other.get_end() >= self.get_begin()

  def insert(self, other: _Chunk):
    """Inserts the other chunk while preserving the begin time of the current chunk"""
    if other.get_begin() is None or other.get_end() is None:
      raise RuntimeError("Cannot insert chunk with no begin time")

    if not self.overlap(other):
      raise RuntimeError("Cannot insert chunk that does not overlap")

    loc = 2 * max(0, other.get_begin() - self.get_begin())
    self._octet_buffer[loc:loc] = other._octet_buffer

    # adjust the begin time preserving the end time
    self.set_begin(self.get_begin() - other.get_dur())

  def __len__(self):
    return len(self._octet_buffer)

  def to_string(self, fps : Fraction, is_df: bool, start_offset: int) -> str:
    if fps not in (FPS_29_97, FPS_30):
      raise ValueError(f"Frame rate {fps} out-of-range")

    if fps != FPS_29_97 and is_df:
      raise ValueError("Frame rate must be fractional if drop frame is true")

    def _octet2hex(octet):
      return format(_Chunk.ODD_PARITY_OCTETS[octet], 'x')

    packets = []
    for i in range(len(self._octet_buffer) // 2):
      packets.append(_octet2hex(self._octet_buffer[2 * i]) + _octet2hex(self._octet_buffer[2 * i + 1]))
    if len(self._octet_buffer) % 2 == 1:
      packets.append(_octet2hex(self._octet_buffer[-1]) + _octet2hex(0))

    return str(SmpteTimeCode.from_frames(self.get_begin() + start_offset, fps, is_df)) + "\t" + " ".join(packets)

MAX_LINEWIDTH = 32

#
# scc writer
#
def from_model(doc: model.ContentDocument, config: Optional[SccWriterConfiguration] = None, progress_callback=lambda _: None) -> str:
  """Converts the data model to an SCC document"""

  # split progress between ISD construction and SCC writing
  def _isd_progress(progress: float):
    progress_callback(progress / 2)

  config : SccWriterConfiguration = config if config is not None else SccWriterConfiguration()
  isds = ISD.generate_isd_sequence(doc, _isd_progress)
  is_rollup = None
  is_last_empty = True

  # generate list of captions
  captions: List[_Caption] = []
  for i, (begin, isd) in enumerate(isds):

    # 25% for creating captions
    progress_callback(0.5 + (i + 1) / len(isds) / 4)
    LOGGER.debug("Processing ISD at %ss to SCC content", float(begin))

    if len(captions) > 0 and captions[-1].get_end() is None:
      captions[-1].set_end(begin)

    non_empty_region_cnt = 0
    for r in isd.iter_regions():
      if len(r) > 0:
        non_empty_region_cnt += 1
        if non_empty_region_cnt > 1:
          # SCC can only handle one region at a time
          LOGGER.warning("Merging multiple regions at %ss; errors may result", float(begin))
          break

    if non_empty_region_cnt == 0:
      # skip empty ISD
      is_last_empty = True
      continue

    is_last_empty = False

    caption: _Caption = _Caption.from_regions(list(isd.iter_regions()))
    caption.set_begin(begin)

    if len(caption) == 0:
      # skip ISD with no text
      continue

    if any(len(e) > MAX_LINEWIDTH for e in caption):
      if config.allow_reflow:
        LOGGER.warning("Line width exceeded at %ss", float(begin))
      else:
        raise RuntimeError(f"Line width exceeded at {float(begin)}s, reflow disabled")
      # merge the lines of the caption and remove duplicate spaces
      text: str = re.sub(' +', ' ', ' '.join(caption))

      # reflow text
      reflowed_lines: List[str] = []
      while len(text) > MAX_LINEWIDTH:
        break_i = MAX_LINEWIDTH

        for i in range(len(text) - 2, 0, -1):
          c = text[i]
          if c == ord(' ') and i < MAX_LINEWIDTH:
            break_i = i
            break

        reflowed_lines.append(text[:break_i])
        text = text[break_i + 1:]

      reflowed_lines.append(text)

      caption.set_lines(reflowed_lines)

    # detect roll-up captions
    if len(captions) > 1 and is_rollup is not False and not is_last_empty:
      if caption[-1].startswith(captions[-1][-1]) or \
        len(caption) > 1 and caption[-2] == captions[-1][-1]:
        is_rollup = True
      else:
        if is_rollup is True:
          LOGGER.warning("Inconsistent roll-up captions, defaulting to pop-on")
        is_rollup = False

    captions.append(caption)

  chunks : List[_Chunk] = []
  for i, caption in enumerate(captions):
    # 25% for SCC writing
    progress_callback(0.75 + (i + 1) / len(captions) / 4)

    if is_rollup is True and not config.force_popon:
      ru_chunk: _Chunk = _Chunk()

      is_painton = i > 0 and caption[-1].startswith(captions[i - 1][-1])

      begin_f = int(caption.get_begin() * config.frame_rate.fps)

      if not is_painton:
        if config.rollup_lines == 2:
          ru_chunk.push_control_code(SccControlCode.RU2.get_ch1_value())
        elif config.rollup_lines == 3:
          ru_chunk.push_control_code(SccControlCode.RU3.get_ch1_value())
        else:
          ru_chunk.push_control_code(SccControlCode.RU4.get_ch1_value())
        # the caption begins when the CR code is received
        begin_f = begin_f - ru_chunk.get_dur()
        ru_chunk.push_control_code(SccControlCode.CR.get_ch1_value())
        pac = SccPreambleAddressCode(1, 15, NamedColors.white, 0, False, False)
        ru_chunk.push_control_code(pac.get_ch1_packet())


      if len(chunks) > 0 and begin_f < chunks[-1].get_end():
        begin_f = chunks[-1].get_end()
        LOGGER.warning("Overlapping roll-up text at %s", SmpteTimeCode.from_seconds(caption.get_begin(), config.frame_rate.fps))
      ru_chunk.set_begin(begin_f)

      for c in (caption[-1][len(captions[i - 1][-1]):] if is_painton else caption[-1]):
        ru_chunk.push_char(c)

      chunks.append(ru_chunk)

      # erase the display if there is a gap between roll-up captions
      if caption.get_end() is not None and \
        (i == len(captions) - 1 or caption.get_end() != captions[i + 1].get_begin()):
        edm_chunk = _Chunk()
        edm_chunk.push_control_code(SccControlCode.EDM.get_ch1_value())
        edm_chunk.set_begin(int(caption.get_end() * config.frame_rate.fps))
        chunks.append(edm_chunk)

    else:
      enm_chunk: _Chunk = _Chunk()

      enm_chunk.push_control_code(SccControlCode.RCL.get_ch1_value())
      enm_chunk.push_control_code(SccControlCode.ENM.get_ch1_value())
      for line_num, line in enumerate(caption, 15 - len(caption)):
        if caption.get_alignment() == TextAlignType.center:
          indent = (32 - len(line)) // 2
        elif caption.get_alignment() == TextAlignType.end:
          indent = 32 - len(line)
        else:
          indent = None

        spaces = indent % 4 if indent is not None else 0
        indent = 4 * (indent // 4) if indent is not None else None

        pac = SccPreambleAddressCode(1, line_num, NamedColors.white, indent, False, False)
        enm_chunk.push_control_code(pac.get_ch1_packet())

        for i in range(spaces):
          enm_chunk.push_char(" ")
        for c in line:
          enm_chunk.push_char(c)
      enm_chunk.push_control_code(SccControlCode.EOC.get_ch1_value())

      enm_chunk.set_begin(int(caption.get_begin() * config.frame_rate.fps - enm_chunk.get_dur() + 2))
      # check if there is an overlap with the previous chunk
      if len(chunks) > 0:
        if chunks[-2].get_end() + chunks[-1].get_dur() > enm_chunk.get_begin():
          LOGGER.warning("Skipping ISD at %s due to overlap in line 21 packets with previous ISD", float(caption.get_begin()))
          continue

        if enm_chunk.overlap(chunks[-1]):
          enm_chunk.insert(chunks[-1])
          chunks.pop()

      chunks.append(enm_chunk)

      # initialize the EDM chunk
      if caption.get_end() is not None:
        edm_chunk = _Chunk()
        edm_chunk.push_control_code(SccControlCode.EDM.get_ch1_value())
        edm_chunk.set_begin(int(caption.get_end() * config.frame_rate.fps))
        chunks.append(edm_chunk)

  start_offset = 0
  if config.start_tc is not None:
    start_tc = SmpteTimeCode.parse(config.start_tc, config.frame_rate.fps)
    if start_tc.is_drop_frame() != config.frame_rate.df:
      raise RuntimeError("The drop-frame status of the specified start_timecode does not match the drop-frame status of the specified frame_rate")
    start_offset = start_tc.to_frames()

  for chunk in chunks:
    if start_offset + chunk.get_begin() < 0:
      raise RuntimeError("The SCC stream would start earlier than the specified start timecode")

  return "Scenarist_SCC V1.0\n\n" + "\n\n".join(map(lambda e: e.to_string(config.frame_rate.fps, config.frame_rate.df, start_offset), chunks))



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC configuration"""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from fractions import Fraction
import typing

from ttconv.config import ModuleConfiguration
from ttconv.style_properties import TextAlignType
from ttconv.time_code import SmpteTimeCode


class TextAlignment(Enum):
  """Text alignment descriptions"""

  LEFT = ("left", TextAlignType.start)
  CENTER = ("center", TextAlignType.center)
  RIGHT = ("right", TextAlignType.end)
  AUTO = ("auto", TextAlignType.start)

  def __init__(self, label: str, text_align: TextAlignType):
    self.label = label
    self.text_align = text_align

  @staticmethod
  def from_value(value: [str, TextAlignment]) -> TextAlignment:
    """Try to create a text alignment instance from the specified value"""
    if isinstance(value, TextAlignment):
      return value

    for text_alignment in list(TextAlignment):
      if value.lower() == text_alignment.label.lower():
        return text_alignment

    raise ValueError(f"Invalid text align '{value}' value. Expect: 'left', 'center', 'right' or 'auto'.")

class SCCFrameRate(Enum):
  """SCC Frame Rates"""

  FPS_30_NDF = ("30NDF", Fraction(30), False)
  FPS_2997_NDF = ("29.97NDF", Fraction(30000, 1001), False)
  FPS_2997_DF = ("29.97DF", Fraction(30000, 1001), True)

  def __init__(self, label: str, fps: Fraction, is_df: bool):
    self.label : str = label
    self.fps : Fraction = fps
    self.df : bool = is_df

  @staticmethod
  def from_value(value: str | SCCFrameRate) -> SCCFrameRate:
    """Create a Frame Rate from a specified value"""
    if isinstance(value, SCCFrameRate):
      return value

    if isinstance(value, str):
      for e in list(SCCFrameRate):
        if value.lower() == e.label.lower():
          return e

    raise ValueError(f"Invalid SCC Frame Rate '{value}'. Expect {','.join([e.label for e in list(SCCFrameRate)])}.")

@dataclass
class SccReaderConfiguration(ModuleConfiguration):
  """SCC reader configuration"""

  text_align: TextAlignment = field(
    default=TextAlignment.AUTO,
    metadata={"decoder": TextAlignment.from_value}
  )

  @classmethod
  def name(cls):
    return "scc_reader"

def _decode_rollup_lines(value: str) -> int:
  decoded_value = int(value)

  if decoded_value < 2 or decoded_value > 4:
    raise ValueError(f"Invalid rollup_lines '{value}' value. Expect: 2-4.")

  return decoded_value

@dataclass
class SccWriterConfiguration(ModuleConfiguration):
  """SCC writer configuration"""

  allow_reflow: bool = field(
    default=True,
    metadata={"decoder": bool}
  )

  force_popon: bool = field(
    default=False,
    metadata={"decoder": bool}
  )

  rollup_lines: int = field(
    default=4,
    metadata={"decoder": _decode_rollup_lines}
  )

  frame_rate: SCCFrameRate = field(
    default=SCCFrameRate.FPS_2997_DF,
    metadata={"decoder": SCCFrameRate.from_value}
  )

  start_tc: typing.Optional[str] = field(
    default=None,
    metadata={"decoder": lambda y: str(y) if y is not None else None}
  )

  @classmethod
  def name(cls):
    return "scc_writer"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""SCC reader"""

from __future__ import annotations

import logging
from typing import Optional

from ttconv.model import ContentDocument, Body, Div, CellResolutionType, ActiveAreaType
from ttconv.scc.caption_paragraph import SCC_SAFE_AREA_CELL_RESOLUTION_ROWS, \
  SCC_SAFE_AREA_CELL_RESOLUTION_COLUMNS, SCC_ROOT_CELL_RESOLUTION_ROWS, SCC_ROOT_CELL_RESOLUTION_COLUMNS
from ttconv.scc.config import SccReaderConfiguration
from ttconv.scc.context import SccContext
from ttconv.scc.line import SccLine
from ttconv.style_properties import StyleProperties, LengthType, GenericFontFamilyType

LOGGER = logging.getLogger(__name__)


#
# SCC reader
#

def to_model(scc_content: str, config: Optional[SccReaderConfiguration] = None, progress_callback=lambda _: None):
  """Converts a SCC document to the data model"""

  document = ContentDocument()

  # Safe area must be a 32x15 grid, that represents 80% of the root area
  root_cell_resolution = CellResolutionType(rows=SCC_ROOT_CELL_RESOLUTION_ROWS, columns=SCC_ROOT_CELL_RESOLUTION_COLUMNS)
  document.set_cell_resolution(root_cell_resolution)

  safe_area_x_offset = int((root_cell_resolution.columns - SCC_SAFE_AREA_CELL_RESOLUTION_COLUMNS) / 2)
  safe_area_y_offset = int((root_cell_resolution.rows - SCC_SAFE_AREA_CELL_RESOLUTION_ROWS) / 2)

  context = SccContext(safe_area_x_offset, safe_area_y_offset, config)

  # The active area is equivalent to the safe area
  active_area = ActiveAreaType(
    left_offset=context.safe_area_x_offset / root_cell_resolution.columns,
    top_offset=context.safe_area_y_offset / root_cell_resolution.rows,
    width=(root_cell_resolution.columns - (context.safe_area_x_offset * 2)) / root_cell_resolution.columns,
    height=(root_cell_resolution.rows - (context.safe_area_y_offset * 2)) / root_cell_resolution.rows,
  )
  document.set_active_area(active_area)

  body = Body()
  body.set_doc(document)
  document.set_body(body)

  # the default value of LineHeight ("normal") typically translates to 125% of the font size, which causes regions to overflow.
  body.set_style(StyleProperties.LineHeight, LengthType(value=100, units=LengthType.Units.pct))

  # use a more readable font than the default Courier
  body.set_style(StyleProperties.FontFamily, ("Consolas", "Monaco", GenericFontFamilyType.monospace))

  # add line padding
  body.set_style(StyleProperties.LinePadding, LengthType(value=0.25, units=LengthType.Units.c))

  context.div = Div()
  context.div.set_doc(document)
  body.push_child(context.div)

  lines = scc_content.splitlines()
  nb_lines = len(lines)

  for (index, line) in enumerate(lines):
    LOGGER.debug(line)
    scc_line = SccLine.from_str(line)

    progress_callback((index + 1) / nb_lines)

    if scc_line is None:
      continue

    scc_line.process(context)

  context.flush()

  return document


def to_disassembly(scc_content: str, show_channels = False) -> str:
  """Dumps an SCC document into the disassembly format"""
  disassembly = ""
  for line in scc_content.splitlines():
    LOGGER.debug(line)
    scc_line = SccLine.from_str(line)

    if scc_line is None:
      continue

    line_to_disassembly = scc_line.to_disassembly(show_channels)
    LOGGER.debug(line_to_disassembly)

    disassembly += line_to_disassembly + "\n"

  return disassembly



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Unit tests for the regions merging filter"""

# pylint: disable=R0201,C0115,C0116,W0212

import unittest

from ttconv.filters.isd.merge_regions import RegionsMergingISDFilter
from ttconv.isd import ISD
from ttconv.model import P, Body, Div, Span, Text, ContentElement


class RegionsMergingFilterTest(unittest.TestCase):

  @staticmethod
  def _get_filled_body(isd: ISD, text: str) -> Body:
    body = Body(isd)
    text = Text(isd, text)
    span = Span(isd)
    span.push_child(text)

    p = P(isd)
    p.push_child(span)

    div = Div(isd)
    div.push_child(p)

    body.push_child(div)

    return body

  @staticmethod
  def _get_text_from_children(element: ContentElement) -> str:
    if isinstance(element, Text):
      return element.get_text()

    for child in element:
      return RegionsMergingFilterTest._get_text_from_children(child)

  def test_merging_regions(self):
    regions_merging_filter = RegionsMergingISDFilter()

    isd = ISD(None)

    r1 = ISD.Region("r1", isd)
    b1 = self._get_filled_body(isd, "Hello world")
    r1.push_child(b1)

    r2 = ISD.Region("r2", isd)
    b2 = self._get_filled_body(isd, "Is there anyone here?")
    r2.push_child(b2)

    isd.put_region(r1)
    isd.put_region(r2)

    self.assertEqual(2, len(list(isd.iter_regions())))

    regions_merging_filter.process(isd)

    self.assertEqual(1, len(list(isd.iter_regions())))

    merged_region = isd.get_region("r1_r2")
    self.assertIsNotNone(merged_region)

    body = list(merged_region)
    self.assertEqual(1, len(body))

    divs = list(body[0])
    self.assertEqual(2, len(divs))

    text = self._get_text_from_children(divs[0])
    self.assertEqual("Hello world", text)

    text = self._get_text_from_children(divs[1])
    self.assertEqual("Is there anyone here?", text)

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

# Copyright (c) 2020, Sandflow Consulting LLC
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Unit tests for the data model'''

# pylint: disable=R0201,C0115,C0116

import unittest
import ttconv.model as model
import ttconv.style_properties as styles
from fractions import Fraction

class ContentElementTest(unittest.TestCase):

  def test_init(self):
    e = model.ContentElement()

    self.assertTrue(e.root() is e)
    self.assertIsNone(e.parent())
    self.assertFalse(e.has_children())
    self.assertIsNone(e.get_doc())

  def test_attach(self):
    doc = model.ContentDocument()

    e = model.ContentElement()

    e.set_doc(doc)

    self.assertTrue(e.get_doc() is doc)
    self.assertTrue(e.is_attached())

  def test_already_attached(self):
    doc = model.ContentDocument()
    doc2 = model.ContentDocument()

    e = model.ContentElement()

    e.set_doc(doc)

    with self.assertRaises(RuntimeError):
      e.set_doc(doc2)

  def test_detach(self):
    doc = model.ContentDocument()

    e = model.ContentElement()

    e.set_doc(doc)

    e.set_doc(None)

    self.assertIsNone(e.get_doc())
    self.assertFalse(e.is_attached())

  def test_failed_child_detach(self):
    doc = model.ContentDocument()

    e = model.ContentElement(doc)

    c = model.ContentElement(doc)

    e.push_child(c)

    with self.assertRaises(RuntimeError):
      c.set_doc(None)

  def test_push_child(self):
    p = model.ContentElement()

    c = model.ContentElement()

    p.push_child(c)

    self.assertEqual(p, c.root())

    self.assertTrue(p.has_children())

    self.assertEqual(sum(1 for _ in p), 1)

    for a_child in p:
      self.assertEqual(a_child, c)

  def test_push_child_with_existing_parent(self):
    p = model.ContentElement()

    c1 = model.ContentElement()

    c2 = model.ContentElement()

    c3 = model.ContentElement()

    p.push_child(c1)

    p.push_child(c2)

    c1.push_child(c3)

    with self.assertRaises(RuntimeError):
      p.push_child(c3)

  def test_push_child_as_self(self):
    p = model.ContentElement()

    with self.assertRaises(RuntimeError):
      p.push_child(p)

  def test_dfs(self):
    p = model.ContentElement()

    c1 = model.ContentElement()

    c2 = model.ContentElement()

    c3 = model.ContentElement()

    p.push_child(c1)

    p.push_child(c2)

    c1.push_child(c3)

    dfs = list(p.dfs_iterator())

    self.assertListEqual([p, c1, c3, c2], dfs)

  def test_first_last_child(self):
    p = model.ContentElement()

    self.assertIsNone(p.first_child())

    self.assertIsNone(p.last_child())

    c1 = model.ContentElement()

    c2 = model.ContentElement()

    c3 = model.ContentElement()

    p.push_child(c1)

    p.push_child(c2)

    p.push_child(c3)

    self.assertEqual(p.first_child(), c1)

    self.assertEqual(p.last_child(), c3)

  def test_remove(self):
    p = model.ContentElement()

    c1 = model.ContentElement()

    c2 = model.ContentElement()

    p.push_child(c1)

    p.push_child(c2)

    self.assertEqual(sum(1 for _ in p), 2)

    c2.remove()

    self.assertIsNone(c2.parent())

    self.assertIsNone(c2.next_sibling())

    self.assertIsNone(c2.previous_sibling())

    self.assertEqual(sum(1 for _ in p), 1)

    for a_child in p:
      self.assertEqual(a_child, c1)

  def test_remove_child(self):
    p = model.ContentElement()

    c1 = model.ContentElement()

    c2 = model.ContentElement()

    c3 = model.ContentElement()

    p.push_child(c1)

    p.push_child(c2)

    p.push_child(c3)

    c1.remove()

    self.assertEqual(sum(1 for _ in p), 2)

    c3.remove()

    self.assertEqual(sum(1 for _ in p), 1)

    c2.remove()

    self.assertEqual(sum(1 for _ in p), 0)

  def test_remove_children(self):
    p = model.ContentElement()

    c1 = model.ContentElement()

    c2 = model.ContentElement()

    c3 = model.ContentElement()

    p.push_child(c1)

    p.push_child(c2)

    p.push_child(c3)

    p.remove_children()

    self.assertFalse(p.has_children())

  def test_children_accessors(self):
    p = model.ContentElement()

    c1 = model.ContentElement()

    c2 = model.ContentElement()

    c3 = model.ContentElement()

    p.push_child(c1)

    p.push_child(c2)

    p.push_child(c3)

    self.assertEqual(len(p), 3)

    self.assertEqual(p[0], c1)

    self.assertRaises(IndexError, lambda key: p[key], 5)

  def test_id(self):
    p = model.ContentElement()

    self.assertIsNone(p.get_id())

    # good id
    good_id = "a"
    p.set_id(good_id)
    self.assertEqual(p.get_id(), good_id)

    # None
    p.set_id(None)
    self.assertIsNone(p.get_id())

    # bad id
    with self.assertRaises(TypeError):
      bad_id = " "
      p.set_id(bad_id)

  def test_animation_step_init(self):

    model.DiscreteAnimationStep(styles.StyleProperties.Color, 0, 1, styles.NamedColors.aqua.value)

    with self.assertRaises(ValueError):
      model.DiscreteAnimationStep(None, 0, 1, styles.NamedColors.aqua.value)
      
    with self.assertRaises(ValueError):
      model.DiscreteAnimationStep(styles.StyleProperties.Color, 0, 1, None)

    with self.assertRaises(ValueError):
      model.DiscreteAnimationStep(styles.StyleProperties.Color, 0, 1, "hello")

  def test_add_remove_animation_step(self):
    p = model.ContentElement()

    s1 = model.DiscreteAnimationStep(styles.StyleProperties.Color, 0, 1, styles.NamedColors.aqua.value)

    s2 = model.DiscreteAnimationStep(styles.StyleProperties.Color, 2, None, styles.NamedColors.black.value)

    s3 = model.DiscreteAnimationStep(styles.StyleProperties.TextAlign, 1, None, styles.TextAlignType.center)

    p.add_animation_step(s1)

    p.add_animation_step(s2)

    p.add_animation_step(s3)

    self.assertListEqual([s1, s2, s3], list(p.iter_animation_steps()))

    p.remove_animation_step(s2)

    self.assertListEqual([s1, s3], list(p.iter_animation_steps()))

  def test_lang(self):
    p = model.ContentElement()

    self.assertEqual(p.get_lang(), "")

    lang = "fr-CA"

    p.set_lang(lang)

    self.assertEqual(p.get_lang(), lang)

  def test_space(self):
    p = model.ContentElement()

    self.assertEqual(p.get_space(), model.WhiteSpaceHandling.DEFAULT)

    space = model.WhiteSpaceHandling.PRESERVE

    p.set_space(space)

    self.assertEqual(p.get_space(), space)

  def test_begin(self):
    p = model.ContentElement()

    self.assertIsNone(p.get_begin())

    f = Fraction(1, 2)

    p.set_begin(f)

    self.assertEqual(p.get_begin(), f)

  def test_end(self):
    p = model.ContentElement()

    self.assertIsNone(p.get_end())

    f = Fraction(1, 3)

    p.set_end(f)

    self.assertEqual(p.get_end(), f)

  def test_copy_to(self):
    src = model.ContentElement()

    src.set_begin(Fraction(1, 2))
    src.set_end(Fraction(1, 3))
    src.set_lang("fr")
    src.set_space(model.WhiteSpaceHandling.PRESERVE)
    src.set_id("hello")

    src.add_animation_step(model.DiscreteAnimationStep(styles.StyleProperties.Color, 0, 1, styles.NamedColors.aqua.value))
    src.set_style(styles.StyleProperties.Color, styles.NamedColors.green.value)

    dest = model.ContentElement()

    src.copy_to(dest)

    self.assertEqual(dest.get_begin(), src.get_begin())
    self.assertEqual(dest.get_end(), src.get_end())
    self.assertEqual(dest.get_lang(), src.get_lang())
    self.assertEqual(dest.get_space(), src.get_space())
    self.assertEqual(dest.get_id(), src.get_id())

    self.assertSequenceEqual(list(dest.iter_animation_steps()), list(src.iter_animation_steps()))

    self.assertSequenceEqual(list(dest.iter_styles()), list(src.iter_styles()))

class BodyTest(unittest.TestCase):

  def test_push_child(self):
    b = model.Body()

    b.push_child(model.Div())

    with self.assertRaises(TypeError):
      b.push_child(model.P())

class DivTest(unittest.TestCase):

  def test_push_child(self):

    d1 = model.Div()

    d1.push_child(model.P())

    d1.push_child(model.Div())

    with self.assertRaises(TypeError):
      d1.push_child(model.Text())

class PTest(unittest.TestCase):

  def test_push_child(self):

    p = model.P()

    p.push_child(model.Br())

    p.push_child(model.Span())

    p.push_child(model.Ruby())

    with self.assertRaises(TypeError):
      p.push_child(model.Text())


class SpanTest(unittest.TestCase):

  def test_push_child(self):

    s = model.Span()

    s.push_child(model.Br())

    s.push_child(model.Span())

    s.push_child(model.Text())

    with self.assertRaises(TypeError):
      s.push_child(model.P())

class BrTest(unittest.TestCase):

  def test_copy_to(self):
    src = model.Br()

    src.set_lang("fr")
    src.set_space(model.WhiteSpaceHandling.PRESERVE)
    src.set_id("hello")

    src.add_animation_step(model.DiscreteAnimationStep(styles.StyleProperties.Color, 0, 1, styles.NamedColors.aqua.value))
    src.set_style(styles.StyleProperties.Color, styles.NamedColors.green.value)

    dest = model.Br()

    src.copy_to(dest)

    self.assertEqual(dest.get_lang(), src.get_lang())
    self.assertEqual(dest.get_space(), src.get_space())
    self.assertEqual(dest.get_id(), src.get_id())

    self.assertSequenceEqual(list(dest.iter_animation_steps()), list(src.iter_animation_steps()))

    self.assertSequenceEqual(list(dest.iter_styles()), list(src.iter_styles()))


class TextTest(unittest.TestCase):

  def test_push_child(self):

    t = model.Text()

    with self.assertRaises(RuntimeError):
      t.push_child(model.Span())

  def test_set_text(self):

    t = model.Text()

    self.assertEqual(t.get_text(), "")

    sample_text = "hello"

    t.set_text(sample_text)

    self.assertEqual(t.get_text(), sample_text)

    with self.assertRaises(TypeError):
      t.set_text(None)

  def test_copy_to(self):
    src = model.Text()

    src.set_text("hello")

    dest = model.Text()

    src.copy_to(dest)

    self.assertEqual(dest.get_text(), src.get_text())

class RegionTest(unittest.TestCase):

  def test_push_child(self):

    r = model.Region("hello")

    with self.assertRaises(RuntimeError):
      r.push_child(model.P())

  def test_set_id(self):

    r = model.Region("hello")

    with self.assertRaises(RuntimeError):
      r.set_id("blah")

  def test_set_region(self):

    r1 = model.Region("hello")

    r2 = model.Region("blah")

    with self.assertRaises(RuntimeError):
      r1.set_region(r2)

  def test_copy_to(self):
    src = model.Region("hello")

    src.set_begin(Fraction(1, 2))
    src.set_end(Fraction(1, 3))
    src.set_lang("fr")
    src.set_space(model.WhiteSpaceHandling.PRESERVE)

    src.add_animation_step(model.DiscreteAnimationStep(styles.StyleProperties.Color, 0, 1, styles.NamedColors.aqua.value))
    src.set_style(styles.StyleProperties.Color, styles.NamedColors.green.value)

    dest = model.Region("hello")

    src.copy_to(dest)

    self.assertEqual(dest.get_begin(), src.get_begin())
    self.assertEqual(dest.get_end(), src.get_end())
    self.assertEqual(dest.get_lang(), src.get_lang())
    self.assertEqual(dest.get_space(), src.get_space())

    self.assertSequenceEqual(list(dest.iter_animation_steps()), list(src.iter_animation_steps()))

    self.assertSequenceEqual(list(dest.iter_styles()), list(src.iter_styles()))

class RubyTest(unittest.TestCase):

  def test_push_child(self):

    r = model.Ruby()

    with self.assertRaises(RuntimeError):
      r.push_child(model.P())

  def test_remove_child(self):

    r = model.Ruby()

    c1 = model.Rb()

    c2 = model.Rt()

    r.push_children([c1, c2])

    with self.assertRaises(RuntimeError):
      c1.remove()

    with self.assertRaises(RuntimeError):
      r.remove_child(c1)

  def test_push_children(self):

    r = model.Ruby()

    r.push_children([model.Rb(), model.Rt()])

    r.remove_children()

    r.push_children([model.Rb(), model.Rp(), model.Rt(), model.Rp()])

    r.remove_children()

    r.push_children([model.Rbc(), model.Rtc()])

    r.remove_children()

    r.push_children([model.Rbc(), model.Rtc(), model.Rtc()])

    with self.assertRaises(RuntimeError):
      r.push_children([model.Rb(), model.Rt()])

class RtcTest(unittest.TestCase):

  def test_push_child_bad(self):

    r = model.Rtc()

    with self.assertRaises(ValueError):
      r.push_child(model.P())

  def test_remove_child(self):

    r = model.Rtc()

    c1 = model.Rt()

    c2 = model.Rt()

    r.push_children([c1, c2])

    with self.assertRaises(RuntimeError):
      c1.remove()

    with self.assertRaises(RuntimeError):
      r.remove_child(c1)

  def test_push_children(self):

    r = model.Rtc()

    r.push_children([model.Rt(), model.Rt()])

    r.remove_children()

    r.push_children([model.Rp(), model.Rt(), model.Rt(), model.Rp()])

    with self.assertRaises(ValueError):
      r.push_children([model.Rb(), model.Rt()])

  def test_push_child(self):

    r = model.Rtc()

    r.push_child(model.Rt())
    r.push_child(model.Rt())

    r.remove_children()

    r.push_children([model.Rp(), model.Rt(), model.Rt(), model.Rp()])

    with self.assertRaises(ValueError):
      r.push_child(model.Rt())

class RbTest(unittest.TestCase):

  def test_push_child(self):

    s = model.Rb()

    s.push_child(model.Span())

    with self.assertRaises(TypeError):
      s.push_child(model.P())


class RtTest(unittest.TestCase):

  def test_push_child(self):

    s = model.Rt()

    s.push_child(model.Span())

    with self.assertRaises(TypeError):
      s.push_child(model.P())


class RpTest(unittest.TestCase):

  def test_push_child(self):

    s = model.Rp()

    s.push_child(model.Span())

    with self.assertRaises(TypeError):
      s.push_child(model.P())


class RbcTest(unittest.TestCase):

  def test_push_child(self):

    s = model.Rbc()

    s.push_child(model.Rb())

    with self.assertRaises(TypeError):
      s.push_child(model.P())


if __name__ == '__main__':
  unittest.main()
